# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""
**routingpy** is a Python 3 client for a lot of popular web routing services.

Using **routingpy** you can easily request **directions**, **isochrones** and
**matrices** from many reliable online providers in a consistent fashion. Base parameters
are the same for all services, while still preserving each service's special parameters (for
more info, please look at our `README`_.
Take a look at our `Examples`_ to see how simple you can compare routes from different providers.

**routingpy** is tested against 3.5, 3.6, 3.7 and PyPy3.

.. _`README`: https://github.com/gis-ops/routing-py#api
.. _`Examples`: https://github.com/gis-ops/routing-py#examples
"""

from routingpy.routers import *

# Delete so options is only available over routingpy.routers.options
del options



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""Converts Python types to string representations suitable for GET queries.
"""


def _delimit_list(arg, delimiter=','):
    """Convert list to delimiter-separated string"""
    if not _is_list(arg):
        raise TypeError("Expected a list or tuple, " "but got {}".format(type(arg).__name__))
    return delimiter.join(map(str, arg))


def _convert_bool(boolean):
    """Convert to stringified boolean"""

    return str(boolean).lower()


def _format_float(arg):
    """Formats a float value to be as short as possible.

    Trims extraneous trailing zeros and period to give API
    args the best possible chance of fitting within 2000 char
    URL length restrictions.

    For example:

    format_float(40) -> "40"
    format_float(40.0) -> "40"
    format_float(40.1) -> "40.1"
    format_float(40.001) -> "40.001"
    format_float(40.0010) -> "40.001"

    :param arg: The lat or lng float.
    :type arg: float

    :rtype: string
    """
    return ("{}".format(round(float(arg), 6)).rstrip("0").rstrip("."))


def _is_list(arg):
    """Checks if arg is list-like."""
    if isinstance(arg, dict):
        return False
    if isinstance(arg, str):  # Python 3-only, as str has __iter__
        return False
    return (
        not _has_method(arg, "strip") and _has_method(arg, "__getitem__")
        or _has_method(arg, "__iter__")
    )


def _has_method(arg, method):
    """Returns true if the given object has a method with the given name.

    :param arg: the object

    :param method: the method name
    :type method: string

    :rtype: bool
    """
    return hasattr(arg, method) and callable(getattr(arg, method))



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

from .base import Router, DEFAULT
from routingpy import convert
from routingpy import utils
from routingpy.direction import Direction, Directions
from routingpy.isochrone import Isochrone, Isochrones
from routingpy.matrix import Matrix


class Graphhopper(Router):
    """Performs requests to the Graphhopper API services."""

    _DEFAULT_BASE_URL = "https://graphhopper.com/api/1"

    def __init__(
            self,
            api_key=None,
            base_url=_DEFAULT_BASE_URL,
            user_agent=None,
            timeout=DEFAULT,
            retry_timeout=None,
            requests_kwargs={},
            retry_over_query_limit=False,
            skip_api_error=None
    ):
        """
        Initializes an graphhopper client.

        :param api_key: GH API key. Required if https://graphhopper.com/api is used.
        :type api_key: str

        :param base_url: The base URL for the request. Defaults to the ORS API
            server. Should not have a trailing slash.
        :type base_url: str

        :param user_agent: User Agent to be used when requesting.
            Default :attr:`routingpy.routers.options.default_user_agent`.
        :type user_agent: str

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify ``None`` for no timeout. Default :attr:`routingpy.routers.options.default_timeout`.
        :type timeout: int or None

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.  Default :attr:`routingpy.routers.options.default_retry_timeout`.
        :type retry_timeout: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. **Note**, that ``proxies`` can be set globally
            in :attr:`routingpy.routers.options.default_proxies`.

            Example:

            >>> from routingpy.routers import Graphhopper
            >>> router = Graphhopper(my_key, requests_kwargs={'proxies': {'https': '129.125.12.0'}})
            >>> print(router.proxies)
            {'https': '129.125.12.0'}
        :type requests_kwargs: dict

        :param retry_over_query_limit: If True, client will not raise an exception
            on HTTP 429, but instead jitter a sleeping timer to pause between
            requests until HTTP 200 or retry_timeout is reached.
            Default :attr:`routingpy.routers.options.default_over_query_limit`.
        :type retry_over_query_limit: bool

        :param skip_api_error: Continue with batch processing if a :class:`routingpy.exceptions.RouterApiError` is
            encountered (e.g. no route found). If False, processing will discontinue and raise an error.
            Default :attr:`routingpy.routers.options.default_skip_api_error`.
        :type skip_api_error: bool
        """

        if base_url == self._DEFAULT_BASE_URL and api_key is None:
            raise KeyError("API key must be specified.")
        self.key = api_key

        super(Graphhopper, self).__init__(
            base_url, user_agent, timeout, retry_timeout, requests_kwargs, retry_over_query_limit,
            skip_api_error
        )

    def directions(
            self,
            locations,
            profile,
            format=None,
            optimize=None,
            instructions=None,
            locale=None,
            elevation=None,
            points_encoded=None,
            calc_points=None,
            debug=None,
            point_hint=None,
            details=None,
            ch_disable=None,
            weighting=None,
            heading=None,
            heading_penalty=None,
            pass_through=None,
            block_area=None,
            avoid=None,
            algorithm=None,
            round_trip_distance=None,
            round_trip_seed=None,
            alternative_route_max_paths=None,
            alternative_route_max_weight_factor=None,
            alternative_route_max_share_factor=None,
            dry_run=None
    ):
        """Get directions between an origin point and a destination point.

        For more information, visit https://docs.graphhopper.com/#tag/Routing-API/paths/~1route/get.

        :param locations: The coordinates tuple the route should be calculated
            from in order of visit.
        :type locations: list of list or tuple of tuple

        :param profile: The vehicle for which the route should be calculated. One of ["car" "bike" "foot" "hike" "mtb"
            "racingbike" "scooter" "truck" "small_truck"]. Default "car".
        :type profile: str

        :param format: Specifies the resulting format of the route, for json the content type will be application/json.
            Default "json".
        :type format: str

        :param locale: Language for routing instructions. The locale of the resulting turn instructions.
            E.g. pt_PT for Portuguese or de for German. Default "en".
        :type locale: str

        :param optimize: If false the order of the locations will be identical to the order of the point parameters.
            If you have more than 2 points you can set this optimize parameter to ``True`` and the points will be sorted
            regarding the minimum overall time - e.g. suiteable for sightseeing tours or salesman.
            Keep in mind that the location limit of the Route Optimization API applies and the credit costs are higher!
            Note to all customers with a self-hosted license: this parameter is only available if your package includes
            the Route Optimization API. Default False.
        :type optimize: bool

        :param instructions: Specifies whether to return turn-by-turn instructions.
            Default True.
        :type instructions: bool

        :param elevation: If true a third dimension - the elevation - is included in the polyline or in the GeoJson.
            IMPORTANT: If enabled you have to use a modified version of the decoding method or set points_encoded to false.
            See the points_encoded attribute for more details. Additionally a request can fail if the vehicle does not
            support elevation. See the features object for every vehicle.
            Default False.
        :type elevation: bool

        :param points_encoded: If ``False`` the coordinates in point and snapped_waypoints are returned as array using the order
            [lon,lat,elevation] for every point. If true the coordinates will be encoded as string leading to less bandwith usage.
            Default True.
        :type points_encoded: bool

        :param calc_points: If the points for the route should be calculated at all, printing out only distance and time.
            Default True.
        :type calc_points: bool

        :param debug: If ``True``, the output will be formated.
            Default False.
        :type debug: bool

        :param point_hint: Optional parameter. Specifies a hint for each point parameter to prefer a certain street for the
            closest location lookup. E.g. if there is an address or house with two or more neighboring streets you can control
            for which street the closest location is looked up.
        :type point_hint: bool

        :param details: Optional parameter. Optional parameter to retrieve path details. You can request additional details for the
            route: street_name and time. For all motor vehicles we additionally support max_speed, toll (no, all, hgv),
            road_class (motorway, primary, ...), road_environment, and surface. The returned format for one details
            is [fromRef, toRef, value]. The ref references the points of the response. Multiple details are possible
            via multiple key value pairs details=time&details=toll
        :type details: list of str

        :param ch_disable: Always use ch_disable=true in combination with one or more parameters of this table.
            Default False.
        :type ch_disable: bool

        :param weighting: Which kind of 'best' route calculation you need. Other options are shortest
            (e.g. for vehicle=foot or bike) and short_fastest if not only time but also distance is expensive.
            Default "fastest".
        :type weighting: str

        :param heading: Optional parameter. Favour a heading direction for a certain point. Specify either one heading for the start point or as
            many as there are points. In this case headings are associated by their order to the specific points.
            Headings are given as north based clockwise angle between 0 and 360 degree.
        :type heading: list of int

        :param heading_penalty: Optional parameter. Penalty for omitting a specified heading. The penalty corresponds to the accepted time
            delay in seconds in comparison to the route without a heading.
            Default 120.
        :type heading_penalty: int

        :param pass_through: Optional parameter. If true u-turns are avoided at via-points with regard to the heading_penalty.
            Default False.
        :type pass_through: bool

        :param block_area: Optional parameter. Block road access via a point with the format
            latitude,longitude or an area defined by a circle lat,lon,radius or a rectangle lat1,lon1,lat2,lon2.
        :type block_area: str

        :param avoid: Optional semicolon separated parameter. Specify which road classes you would like to avoid
            (currently only supported for motor vehicles like car). Possible values are ferry, motorway, toll, tunnel and ford.
        :type avoid: list of str

        :param algorithm: Optional parameter. round_trip or alternative_route.
        :type algorithm: str

        :param round_trip_distance: If algorithm=round_trip this parameter configures approximative length of the resulting round trip.
            Default 10000.
        :type round_trip_distance: int

        :param round_trip_seed: If algorithm=round_trip this parameter introduces randomness if e.g. the first try wasn't good.
            Default 0.
        :type round_trip_seed: int

        :param alternative_route_max_paths: If algorithm=alternative_route this parameter sets the number of maximum paths
            which should be calculated. Increasing can lead to worse alternatives.
            Default 2.
        :type alternative_route_max_paths: int

        :param alternative_route_max_weight_factor: If algorithm=alternative_route this parameter sets the factor by which the alternatives
            routes can be longer than the optimal route. Increasing can lead to worse alternatives.
            Default 1.4.
        :type alternative_route_max_weight_factor: float

        :param alternative_route_max_share_factor: If algorithm=alternative_route this parameter specifies how much alternatives
            routes can have maximum in common with the optimal route. Increasing can lead to worse alternatives.
            Default 0.6.
        :type alternative_route_max_share_factor: float

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: One or multiple route(s) from provided coordinates and restrictions.
        :rtype: :class:`routingpy.direction.Direction` or :class:`routingpy.direction.Directions`
        """

        params = [('vehicle', profile)]

        for coordinate in locations:
            coord_latlng = reversed([convert._format_float(f) for f in coordinate])
            params.append(("point", ",".join(coord_latlng)))

        if self.key is not None:
            params.append(("key", self.key))

        if format is not None:
            params.append(("type", format))

        if optimize is not None:
            params.append(("optimize", convert._convert_bool(optimize)))

        if instructions is not None:
            params.append(("instructions", convert._convert_bool(instructions)))

        if locale is not None:
            params.append(("locale", locale))

        if elevation is not None:
            params.append(("elevation", convert._convert_bool(elevation)))

        if points_encoded is not None:
            params.append(("points_encoded", convert._convert_bool(points_encoded)))

        if calc_points is not None:
            params.append(("calc_points", convert._convert_bool(calc_points)))

        if debug is not None:
            params.append(("debug", convert._convert_bool(debug)))

        if point_hint is not None:
            params.append(("point_hint", convert._convert_bool(point_hint)))

        ### all below params will only work if ch is disabled

        if details is not None:
            params.extend([("details", detail) for detail in details])

        if ch_disable is not None:
            params.append(("ch.disable", convert._convert_bool(ch_disable)))

        if weighting is not None:
            params.append(("weighting", weighting))

        if heading is not None:
            params.append(("heading", convert._delimit_list(heading)))

        if heading_penalty is not None:
            params.append(("heading_penalty", heading_penalty))

        if pass_through is not None:
            params.append(("pass_through", convert._convert_bool(pass_through)))

        if block_area is not None:
            params.append(("block_area", block_area))

        if avoid is not None:
            params.append(("avoid", convert._delimit_list(avoid, ';')))

        if algorithm is not None:

            params.append(('algorithm', algorithm))

            if algorithm == 'round_trip':

                if round_trip_distance is not None:
                    params.append(("round_trip.distance", round_trip_distance))

                if round_trip_seed is not None:
                    params.append(("round_trip.seed", round_trip_seed))

            if algorithm == 'alternative_route':

                if alternative_route_max_paths is not None:
                    params.append(("alternative_route.max_paths", alternative_route_max_paths))

                if alternative_route_max_weight_factor is not None:
                    params.append(
                        ("alternative_route.max_weight_factor", alternative_route_max_weight_factor)
                    )

                if alternative_route_max_share_factor:
                    params.append(
                        ("alternative_route_max_share_factor", alternative_route_max_share_factor)
                    )

        return self._parse_directions_json(
            self._request('/route', get_params=params, dry_run=dry_run), algorithm, elevation
        )

    @staticmethod
    def _parse_directions_json(response, algorithm, elevation):
        if response is None:  # pragma: no cover
            if algorithm == 'alternative_route':
                return Directions()
            else:
                return Direction()

        if algorithm == 'alternative_route':
            routes = []
            for route in response['paths']:
                geometry = [
                    list(reversed(coord))
                    for coord in utils.decode_polyline5(route['points'], elevation)
                ]
                routes.append(
                    Direction(
                        geometry=geometry,
                        duration=int(route['time'] / 1000),
                        distance=int(route['distance']),
                        raw=route
                    )
                )
            return Directions(routes, response)
        else:
            geometry = [
                list(reversed(coord))
                for coord in utils.decode_polyline5(response['paths'][0]['points'], elevation)
            ]
            return Direction(
                geometry=geometry,
                duration=int(response['paths'][0]['time'] / 1000),
                distance=int(response['paths'][0]['distance']),
                raw=response
            )

    def isochrones(
            self,
            locations,
            profile,
            intervals,
            type='json',
            buckets=1,
            interval_type=None,
            reverse_flow=None,
            debug=None,
            dry_run=None
    ):
        """Gets isochrones or equidistants for a range of time/distance values around a given set of coordinates.

        For mroe details visit https://docs.graphhopper.com/#tag/Isochrone-API.

        :param locations: One coordinate pair denoting the location.
        :type locations: tuple of float or list of float

        :param profile: Specifies the mode of transport.
            One of bike, car, foot or
            https://graphhopper.com/api/1/docs/supported-vehicle-profiles/Default.
            Default "car".
        :type profile: str

        :param intervals: Maximum range to calculate distances/durations for. You can also specify
            the ``buckets`` variable to break the single value into more isochrones. For compatibility reasons,
            this parameter is expressed as list. In meters or seconds.
        :type intervals: list of int or tuple of int

        :param interval_type: Set ``time`` for isochrones or ``distance`` for equidistants.
            Default 'time'.
        :type interval_type: str

        :param buckets: For how many sub intervals an additional polygon should be calculated.
            Default 1.
        :type buckets: int

        :param reverse_flow: If false the flow goes from point to the polygon,
            if true the flow goes from the polygon "inside" to the point.
            Default False.
        :param reverse_flow: bool

        :param debug: If true, the output will be formatted.
            Default False
        :type debug: bool

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: An isochrone with the specified range.
        :rtype: :class:`routingpy.isochrone.Isochrones`
        """

        params = [('vehicle', profile), ('type', type)]

        if convert._is_list(intervals):
            if interval_type in (None, 'time'):
                params.append(('time_limit', intervals[0]))
            elif interval_type == 'distance':
                params.append(('distance_limit', intervals[0]))
        else:
            raise TypeError("Parameter range={} must be of type list or tuple".format(range))

        center = [convert._format_float(f) for f in locations]
        center.reverse()
        params.append(("point", ",".join(center)))

        if self.key is not None:
            params.append(("key", self.key))

        if buckets is not None:
            params.append(('buckets', buckets))

        if reverse_flow is not None:
            params.append(('reverse_flow', convert._convert_bool(reverse_flow)))

        if debug is not None:
            params.append(('debug', convert._convert_bool(debug)))

        return self._parse_isochrone_json(
            self._request("/isochrone", get_params=params, dry_run=dry_run), type, intervals[0], buckets,
            center
        )

    @staticmethod
    def _parse_isochrone_json(response, type, max_range, buckets, center):
        if response is None:  # pragma: no cover
            return Isochrones()

        isochrones = []
        accessor = 'polygons' if type == 'json' else 'features'
        for index, polygon in enumerate(response[accessor]):
            isochrones.append(
                Isochrone(
                    geometry=[l[:2] for l in polygon['geometry']['coordinates'][0]
                              ],  # takes in elevation for some reason
                    interval=int(max_range * ((polygon['properties']['bucket'] + 1) / buckets)),
                    center=center,
                )
            )

        return Isochrones(isochrones, response)

    def matrix(
            self,
            locations,
            profile,
            sources=None,
            destinations=None,
            out_array=['times', 'distances'],
            debug=None,
            dry_run=None
    ):
        """ Gets travel distance and time for a matrix of origins and destinations.

        For more details visit https://docs.graphhopper.com/#tag/Matrix-API.

        :param locations: Specifiy multiple points for which the weight-, route-, time- or distance-matrix should be calculated.
            In this case the starts are identical to the destinations.
            If there are N points, then NxN entries will be calculated.
            The order of the point parameter is important. Specify at least three points.
            Cannot be used together with from_point or to_point. Is a string with the format latitude,longitude.
        :type locations: list of list or tuple of tuple or list of tuple or tuple of list

        :param profile: Specifies the mode of transport.
            One of bike, car, foot or
            https://graphhopper.com/api/1/docs/supported-vehicle-profiles/Default.
            Default "car".
        :type profile: str

        :param sources: The starting points for the routes.
            Specifies an index referring to locations.
        :type sources: list of int

        :param destinations: The destination points for the routes. Specifies an index referring to locations.
        :type destinations: list of int

        :param out_array: Specifies which arrays should be included in the response. Specify one or more of the following
            options 'weights', 'times', 'distances'.
            The units of the entries of distances are meters, of times are seconds and of weights is arbitrary and it can differ
            for different vehicles or versions of this API.
            Default ["times", "distance"].
        :type out_array: list of str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: A matrix from the specified sources and destinations.
        :rtype: :class:`routingpy.matrix.Matrix`
        """
        params = [('vehicle', profile)]

        if self.key is not None:
            params.append(("key", self.key))

        if sources is None and destinations is None:
            locations = (reversed([convert._format_float(f) for f in coord]) for coord in locations)
            params.extend([('point', ",".join(coord)) for coord in locations])

        else:
            sources_out = locations
            destinations_out = locations
            try:
                sources_out = []
                for idx in sources:
                    sources_out.append(locations[idx])
            except IndexError:
                raise IndexError("Parameter sources out of locations range at index {}.".format(idx))
            except TypeError:
                # Raised when sources == None
                pass
            try:
                destinations_out = []
                for idx in destinations:
                    destinations_out.append(locations[idx])
            except IndexError:
                raise IndexError(
                    "Parameter destinations out of locations range at index {}.".format(idx)
                )
            except TypeError:
                # Raised when destinations == None
                pass

            sources_out = (reversed([convert._format_float(f) for f in coord]) for coord in sources_out)
            params.extend([("from_point", ",".join(coord)) for coord in sources_out])

            destinations_out = (
                reversed([convert._format_float(f) for f in coord]) for coord in destinations_out
            )
            params.extend([("to_point", ",".join(coord)) for coord in destinations_out])

        if out_array is not None:
            for e in out_array:
                params.append(("out_array", e))

        if debug is not None:
            params.append(('debug', convert._convert_bool(debug)))

        return self._parse_matrix_json(self._request('/matrix', get_params=params, dry_run=dry_run), )

    @staticmethod
    def _parse_matrix_json(response):
        if response is None:  # pragma: no cover
            return Matrix()
        durations = response.get('times')
        distances = response.get('distances')

        return Matrix(durations=durations, distances=distances, raw=response)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

__title__ = 'routingpy'
__description__ = 'One lib to route them all.'
__url__ = 'https://gis-ops.com/projects/routingpy'
__version__ = '0.2.7'
__author__ = 'Nils Nolde'
__author_email__ = 'nils@gis-ops.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2019 GIS OPS UG'



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

from .base import Router, DEFAULT
from routingpy import convert
from routingpy import utils
from routingpy.direction import Direction, Directions
from routingpy.isochrone import Isochrone, Isochrones
from routingpy.matrix import Matrix


class BingMaps(Router):
    """Performs requests to the Bing Maps API services."""

    _DEFAULT_BASE_URL = "https://graphhopper.com/api/1"



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

from .base import Router, DEFAULT
from routingpy import convert
from routingpy.direction import Direction, Directions
from routingpy.isochrone import Isochrones, Isochrone
from routingpy.matrix import Matrix

from operator import itemgetter


class HereMaps(Router):
    """Performs requests to the HERE Maps API services."""

    _DEFAULT_BASE_URL = 'https://route.api.here.com/routing/7.2'

    def __init__(
            self,
            app_id=None,
            app_code=None,
            user_agent=None,
            timeout=DEFAULT,
            retry_timeout=None,
            requests_kwargs=None,
            retry_over_query_limit=False,
            skip_api_error=None
    ):
        """
        Initializes a HERE Maps client.

        :param app_id: HERE Maps app id. 
        :type app_id: str

        :param app_code: HERE Maps app code.
        :type app_code: str

        :param user_agent: User Agent to be used when requesting.
            Default :attr:`routingpy.routers.options.default_user_agent`.
        :type user_agent: str

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify ``None`` for no timeout. Default :attr:`routingpy.routers.options.default_timeout`.
        :type timeout: int or None

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.  Default :attr:`routingpy.routers.options.default_retry_timeout`.
        :type retry_timeout: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. **Note**, that ``proxies`` can be set globally
            in :attr:`routingpy.routers.options.default_proxies`.

            Example:

            >>> from routingpy.routers import HereMaps
            >>> router = HereMaps(my_api_id, my_api_code, requests_kwargs={'proxies': {'https': '129.125.12.0'}})
            >>> print(router.proxies)
            {'https': '129.125.12.0'}
        :type requests_kwargs: dict

        :param retry_over_query_limit: If True, client will not raise an exception
            on HTTP 429, but instead jitter a sleeping timer to pause between
            requests until HTTP 200 or retry_timeout is reached.
            Default :attr:`routingpy.routers.options.default_over_query_limit`.
        :type retry_over_query_limit: bool

        :param skip_api_error: Continue with batch processing if a :class:`routingpy.exceptions.RouterApiError` is
            encountered (e.g. no route found). If False, processing will discontinue and raise an error.
            Default :attr:`routingpy.routers.options.default_skip_api_error`.
        :type skip_api_error: bool
        """

        if app_id is None and app_code is None:
            raise KeyError("HERE Maps app_id and app_code must be specified.")

        self.app_code = app_code
        self.app_id = app_id

        super(HereMaps, self).__init__(
            self._DEFAULT_BASE_URL, user_agent, timeout, retry_timeout, requests_kwargs,
            retry_over_query_limit, skip_api_error
        )

    class Waypoint(object):
        """
        Constructs a waypoint with additional information.
        https://developer.here.com/documentation/routing/topics/resource-param-type-waypoint.html

        Example:

        >>> waypoint = HereMaps.Waypoint(position=[8.15315, 52.53151], waypoint_type='passThrough', stopover_duration=120, transit_radius=500)
        >>> route = HereMaps(api_key).directions(locations=[[[8.58232, 51.57234]], waypoint, [7.15315, 53.632415]])
        """

        def __init__(
                self,
                position,
                waypoint_type=None,
                stopover_duration=None,
                transit_radius='',
                user_label='',
                heading=''
        ):
            """
            :param position: Indicates that the parameter contains a geographical position.
            :type position: list

            :param waypoint_type: 180 degree turns are allowed for ``stopOver`` but not for ``passThrough``.
                Waypoints defined through a drag-n-drop action should be marked as passThrough.
                PassThrough waypoints will not appear in the list of maneuvers.
            :type waypoint_type: str

            :param stopover_duration: Stopover delay in seconds. 
                Impacts time-aware calculations. Ignored for passThrough.
            :type stopover_duration: int

            :param transit_radius: Matching Links are selected within the 
                specified TransitRadius, in meters. 
                For example to drive past a city without necessarily going into the 
                city center you can specify the coordinates of the center and a 
                TransitRadius of 5000m.
            :type transit_radius: int

            :param user_label: Custom label identifying this waypoint.
            :type user_label: str

            :param heading: Heading in degrees starting at true north and continuing 
                clockwise around the compass. 
                North is 0 degrees, East is 90 degrees, South is 180 degrees, 
                and West is 270 degrees.
            :type heading: int
            """

            self.position = position
            self.waypoint_type = waypoint_type
            self.stopover_duration = stopover_duration
            self.transit_radius = str(transit_radius)
            self.user_label = user_label
            self.heading = str(heading)

        def _make_waypoint(self):

            here_waypoint = ['geo']
            if self.waypoint_type is not None and self.stopover_duration is not None:
                here_waypoint.append(
                    convert._delimit_list([self.waypoint_type, self.stopover_duration], ',')
                )
            elif self.waypoint_type is not None:
                here_waypoint.append(self.waypoint_type)

            position = convert._delimit_list(
                [convert._format_float(f) for f in list(reversed(self.position))], ','
            )
            position += ';' + self.transit_radius
            position += ';' + self.user_label
            position += ';' + self.heading
            here_waypoint.append(position)
            return convert._delimit_list(here_waypoint, '!')

    class RoutingMode(object):
        """
        Optionally construct the routing profile from this class with additional attributes.
        https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html

        Example

        >>> profile = HereMaps.RoutingMode(mode_type='shortest', mode_transport_type='truck', mode_traffic=True, features={'motorway': -2})
        >>> route = HereMaps(api_key).directions(locations=location_list, profile=profile)

        """

        def __init__(
                self, mode_type='fastest', mode_transport_type='car', mode_traffic=None, features=None
        ):
            """
            :param mode_type: RoutingType relevant to calculation.
            :type mode_type: str

            :param mode_transport_type: Specify which mode of transport to calculate the route for.
            :type mode_transport_type: str

            :param mode_traffic: Specify whether to optimize a route for traffic.
            :type mode_traffic: bool

            :param features: Route feature weightings to be applied when calculating 
                the route. As many as required.
            :type features: dict
            """

            self.mode_type = mode_type
            self.mode_transport_type = mode_transport_type
            self.mode_traffic = mode_traffic
            if features is not None:
                self.features = features

        def make_routing_mode(self):

            routing_mode = []
            routing_mode.append(self.mode_type)
            routing_mode.append(self.mode_transport_type)

            if self.mode_traffic is not None:
                routing_mode.append('traffic:' + self.mode_traffic)

            if self.features is not None:
                get_features = []
                for f, w in self.features.items():
                    get_features.append(convert._delimit_list([f, str(w)], ':'))
                routing_mode.append(convert._delimit_list(get_features, ','))
            return convert._delimit_list(routing_mode, ';')

    def directions(
            self,
            locations,
            profile,
            mode_type='fastest',
            format='json',
            request_id=None,
            avoid_areas=None,
            avoid_links=None,
            avoid_seasonal_closures=None,
            avoid_turns=None,
            allowed_zones=None,
            exclude_zones=None,
            exclude_zone_types=None,
            exclude_countries=None,
            arrival=None,
            departure=None,
            alternatives=None,
            metric_system=None,
            view_bounds=None,
            resolution=None,
            instruction_format=None,
            language=None,
            json_attributes=None,
            json_callback=None,
            representation=None,
            route_attributes=['waypoints', 'summary', 'shape', 'boundingBox', 'legs'],
            leg_attributes=None,
            maneuver_attributes=None,
            link_attributes=None,
            line_attributes=None,
            generalization_tolerances=None,
            vehicle_type=None,
            license_plate=None,
            max_number_of_changes=None,
            avoid_transport_types=None,
            walk_time_multiplier=None,
            walk_speed=None,
            walk_radius=None,
            combine_change=None,
            truck_type=None,
            trailers_count=None,
            shipped_hazardous_goods=None,
            limited_weight=None,
            weight_per_axle=None,
            height=None,
            width=None,
            length=None,
            tunnel_category=None,
            truck_restriction_penalty=None,
            return_elevation=None,
            consumption_model=None,
            custom_consumption_details=None,
            speed_profile=None,
            dry_run=None
    ):
        """Get directions between an origin point and a destination point.

        For more information, https://developer.here.com/documentation/routing/topics/resource-calculate-route.html.

        :param locations: The coordinates tuple the route should be calculated
            from in order of visit. Can be a list/tuple of [lon, lat] or :class:`HereMaps.Waypoint` instance or a
            combination of those. For further explanation, see
            https://developer.here.com/documentation/routing/topics/resource-param-type-waypoint.html
        :type locations: list of list or list of :class:`HereMaps.Waypoint`

        :param profile: Specifies the routing mode of transport and further options.
            Can be a str and one of [car, pedestrian, carHOV, publicTransport, publicTransportTimeTable, truck, bicycle]
            or :class:`HereMaps.RoutingMode`.
            https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html
        :type profile: str or :class:`HereMaps.RoutingMode`

        :param mode_type: RoutingType relevant to calculation. One of [fastest, shortest, balanced]. Default fastest.
            https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html#ariaid-title2
        :type mode_type: str

        :param format: Currently only "json" supported. 
        :type format: str

        :param request_id: Clients may pass in an arbitrary string to trace request 
            processing through the system. The RequestId is mirrored in the MetaInfo 
            element of the response structure.
        :type request_id: str

        :param avoid_areas: Areas which the route must not cross. Array of BoundingBox. 
            Example with 2 bounding boxes
            https://developer.here.com/documentation/routing/topics/resource-param-type-bounding-box.html
        :type avoid_areas: list of list of list

        :param avoid_links: Links which the route must not cross. The list of LinkIdTypes.
        :type avoid_areas: list of str

        :param avoid_seasonal_closures: The optional avoid seasonal closures boolean 
            flag can be specified to avoid usage of seasonally closed links.
            Examples of seasonally closed links are roads that may be closed during the 
            winter due to weather conditions or ferries that may be out of operation 
            for the season (based on past closure dates).
        :type avoid_seasonal_closures: bool

        :param avoid_turns: List of turn types that the route should avoid. 
            Defaults to empty list. 
            https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html
        :type avoid_turns: str

        :param allowed_zones: Identifiers of zones where routing engine should 
            not take zone restrictions into account (for example in case of a 
            special permission to access a restricted environmental zone).
            https://developer.here.com/documentation/routing/topics/resource-get-routing-zones.html
        :type allowed_zones: list of int

        :param exclude_zones: Identifiers of zones which the route must not cross 
            under any circumstances.
            https://developer.here.com/documentation/routing/topics/resource-get-routing-zones.html
        :type exclude_zones: list of int

        :param exclude_zone_types: List of zone types which the route must not 
            cross under any circumstances.
            https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html
            #resource-type-enumerations__enum-routing-zone-type-type
        :type exclude_zone_types: list of str

        :param exclude_countries: Countries that must be excluded from route calculation.
        :type exclude_zone_types: list of str

        :param departure: Time when travel is expected to start. Traffic speed and 
            incidents are taken into account
            when calculating the route (note that in case of a past
            departure time the historical traffic is limited to one year).  
            You can use now to specify the current time. Specify either departure 
            or arrival, not both. When the optional timezone offset is not 
            specified, the time is assumed to be the local.
            Formatted as iso time, e.g. 2018-07-04T17:00:00+02.
        :type departure: str

        :param arrival: Time when travel is expected to end. Specify either 
            departure or arrival, not both.
            When the optional timezone offset is not specified, the time is assumed to be the local.
            Formatted as iso time, e.g. 2018-07-04T17:00:00+02.
        :type arrival: str        

        :param alternatives: Maximum number of alternative routes that will be 
            calculated and returned. Alternative routes can be unavailable, thus 
            they are not guaranteed to be returned. If at least one via point is used
            in a route request, returning alternative routes is not supported. 
            0 stands for "no alternative routes", i.e. only best route is returned.
        :type alternatives: int

        :param metric_system: Defines the measurement system used in instruction text. 
            When imperial is selected, units used are based on the language specified 
            in the request. Defaults to metric when not specified.
        :type metric_system: str

        :param view_bounds: If the view bounds are given in the request then only 
            route shape points which fit into these bounds will be returned. 
            The route shape beyond the view bounds is reduced to the points which
            are referenced by links, legs or maneuvers.
        :type view_bounds: list or tuple

        :param resolution: Specifies the resolution of the view and a possible snap 
            resolution in meters per pixel in the response. You must specify a whole, positive integer.
            If you specify only one value, then this value defines the view resolution only.
            You can use snap resolution to adjust waypoint links to the resolution of the client display.
            e.g. {'viewresolution': 300,'snapresolution': 300}
        :type resolution: dict

        :param instruction_format: Defines the representation format of the maneuver's instruction text. Html or txt.
        :type instruction_format: str

        :param language: A list of languages for all textual information, 
            the first supported language is used. If there are no matching supported 
            languages the response is an error. Defaults to en-us.
            https://developer.here.com/documentation/routing/topics/resource-param-type-languages.html#languages
        :type language: str

        :param json_attributes: Flag to control JSON output. Combine parameters 
            by adding their values. 
            https://developer.here.com/documentation/routing/topics/resource-param-type-json-representation.html
        :type json_attributes: int

        :param json_callback: Name of a user-defined function used to wrap the JSON response.
        :type json_callback: str

        :param representation: Define which elements are included in the response 
            as part of the data representation
            of the route.
            https://developer.here.com/documentation/routing/topics/resource-param-type-route-representation-options.html#type-route-represenation-mode
        :type representation: list of str

        :param route_attributes: Define which attributes are included in the 
            response as part of the data representation of the route. Defaults to 
            waypoints, summary, legs and additionally lines if publicTransport or
            publicTransportTimeTable mode is used.
            https://developer.here.com/documentation/routing/topics/resource-param-type-route-representation-options.html#type-route-attribute
        :type route_attributes: list of str

        :param leg_attributes: Define which attributes are included in the response 
            as part of the data representation
            of the route legs. Defaults to maneuvers, waypoint, length, travelTime.
            https://developer.here.com/documentation/routing/topics/resource-param-type-route-representation-options.html#type-route-leg-attribute
        :type leg_attributes: list of str

        :param maneuver_attributes: Define which attributes are included in the 
            response as part of the data
            representation of the route maneuvers. Defaults to position, length, travelTime.
            https://developer.here.com/documentation/routing/topics/resource-param-type-route-representation-options.html#type-maneuver-attribute
        :type maneuver_attributes: list of str

        :param link_attributes: Define which attributes are included in the response 
            as part of the data representation of the route links. Defaults to shape, speedLimit.
            https://developer.here.com/documentation/routing/topics/resource-param-type-route-representation-options.html#type-route-link-attribute
        :type link_attributes: list of str

        :param line_attributes: Sequence of attribute keys of the fields that are 
            included in public transport line elements. If not specified, 
            defaults to lineForeground, lineBackground.
            https://developer.here.com/documentation/routing/topics/resource-param-type-route-representation-options.html#type-public-transport-line-attribute
        :type line_attributes: list of str

        :param generalization_tolerances: Specifies the desired tolerances for 
            generalizations of the base route geometry. Tolerances are given in 
            degrees of longitude or latitude on a spherical approximation of the Earth.
            One meter is approximately equal to 0:00001 degrees at typical latitudes.
        :type generalization_tolerances: list of float

        :param param vehicle_type: Specifies type of vehicle engine and average 
            fuel consumption, which can be used to estimate CO2 emission for
            the route summary.
            https://developer.here.com/documentation/routing/topics/resource-param-type-vehicle-type.html
        :type vehicle_type: str

        :param param license_plate: Specifies fragments of vehicle's license plate number. 
            The lastcharacter is currently the only supported fragment type. 
            The license plate parameter enables evaluation of license plate
            based vehicle restrictions like odd/even scheme in Indonesia.
        :type license_plate: str

        :param max_number_of_changes: Restricts number of changes in a public 
            transport route to a given value. The parameter does not filter resulting 
            alternatives. Instead, it affects route calculation so that only
            routes containing at most the given number of changes are considered. 
            The provided value must be between 0 and 10.
        :type max_number_of_changes: int

        :param avoid_transport_types: Public transport types that shall not be included 
            in the response route. Please refer to Enumeration Types for a list of supported values.
            https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html
        :type avoid_transport_types: list of str 

        :param walk_time_multiplier: Allows to prefer or avoid public transport 
            routes with longer walking distances. A value > 1.0 means a slower 
            walking speed and will prefer routes with less walking distance.
            The provided value must be between 0.01 and 100.
        :type walk_time_multiplier: float

        :param walk_speed: Specifies speed which will be used by a service as a 
            walking speed for pedestrian routing (meters per second). 
            This parameter affects pedestrian, publicTransport and publicTransportTimetable modes.
            The provided value must be between 0.5 and 2.
        :type walk_speed: float

        :param walk_radius: Allows the user to specify a maximum distance to the 
            start and end stations of a public transit route. Only valid for 
            publicTransport and publicTransportTimetable routes.
            The provided value must be between 0 and 6000.
        :type walk_radius: int

        :param combine_change: Enables the change maneuver in the route response, 
            which indicates a public transit line change. In the absence of this 
            maneuver, each line change is represented with a pair of subsequent enter
            and leave maneuvers. We recommend enabling combineChange behavior wherever 
            possible, to simplify client-side development.
        :type combine_change: bool

        :param truck_type: Truck routing only, specifies the vehicle type. Defaults to truck.
        :type truck_type: str

        :param trailers_count: Truck routing only, specifies number of trailers 
            pulled by a vehicle. The provided value must be between 0 and 4. 
            Defaults to 0.
        :type trailers_count: int

        :param shipped_hazardous_goods: Truck routing only, list of hazardous 
            materials in the vehicle. Please refer to the enumeration type 
            HazardousGoodTypeType for available values. 
            Note the value allhazardousGoods does not apply to the request parameter.
            https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html#resource-type-enumerations__enum-hazardous-good-type-type
        :type shipped_hazardous_goods: list of str

        :param limited_weight: Truck routing only, vehicle weight including 
            trailers and shipped goods, in tons. The provided value must be between 0 and 1000.
        :type limited_weight: int

        :param weight_per_axle: Truck routing only, vehicle weight per axle in 
            tons. The provided value must be between 0 and 1000.
        :type limited_weight: int

        :param height: Truck routing only, vehicle height in meters.
            The provided value must be between 0 and 50.
        :type height: int

        :param width: Truck routing only, vehicle width in meters. 
            The provided value must be between 0 and 50.
        :type width: int  

        :param length: Truck routing only, vehicle length in meters.
            The provided value must be between 0 and 300.
        :type length: int        

        :param tunnel_category: Truck routing only, specifies the tunnel category 
            to restrict certain route links. The route will pass only through tunnels 
            of a less strict category.
        :type tunnel_category: list of str

        :param truck_restriction_penalty: Truck routing only, specifies the 
            penalty type on violated truck restrictions. Defaults to strict. 
            Refer to the enumeration type TruckRestrictionPenaltyType for 
            details on available values. 
            https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html#resource-type-enumerations__enum-truck-restriction-penalty-type
        :type truck_restriction_penalty: str

        :param return_elevation: If set to true, all shapes inside routing response
            will consist of 3 values instead of 2. Third value will be elevation. 
            If there are no elevation data available for given shape point, 
            elevation will be interpolated from surrounding points. In case 
            there is no elevation data available for any of the shape points, 
            elevation will be 0.0. If jsonattributes=32, elevation cannot be returned.
        :type return_elevation: bool

        :param consumption_model: If you request information on consumption, 
            you must provide a consumption model. The possible values are default 
            and standard. When you specify the value standard, you must provide 
            additional information in the query parameter customconsumptiondetails
        :type consumption_model: str

        :param custom_consumption_details: Provides vehicle specific information 
            for use in the consumption model. This information can include such 
            things as the amount of energy consumed while travelling at a given speed.
            https://developer.here.com/documentation/routing/topics/resource-param-type-custom-consumption-details.html#type-standard
        :type custom_consumption_details: str

        :param speed_profile: Specifies the speed profile variant for a given routing mode. 
            The speed profile affects travel time estimation as well as roads evaluation 
            when computing the fastest route. Note that computed routes might differ depending on a used profile.
            https://developer.here.com/documentation/routing/topics/resource-param-type-speed-profile-type.html
        :type speed_profile: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: One or multiple route(s) from provided coordinates and restrictions.
        :rtype: :class:`routingpy.direction.Direction` or :class:`routingpy.direction.Directions`

        """

        self.base_url = 'https://route.api.here.com/routing/7.2'
        params = {}

        params["app_code"] = self.app_code
        params["app_id"] = self.app_id

        locations = self._build_locations(locations)

        for idx, wp in enumerate(locations):
            wp_index = "waypoint" + str(idx)
            params[wp_index] = wp

        if isinstance(profile, str):
            params["mode"] = mode_type + ';' + profile
        elif isinstance(profile, self.RoutingMode):
            params["mode"] = profile.make_routing_mode()

        if request_id is not None:
            params["requestId"] = request_id

        if avoid_areas is not None:
            params["avoidAreas"] = convert._delimit_list(
                [
                    convert._delimit_list(
                        [
                            convert._delimit_list(
                                [convert._format_float(f) for f in list(reversed(pair))], ','
                            ) for pair in bounding_box
                        ], ';'
                    ) for bounding_box in avoid_areas
                ], '!'
            )

        if avoid_links is not None:
            params["avoidLinks"] = convert._delimit_list(avoid_links, ',')

        if avoid_seasonal_closures is not None:
            params["avoidSeasonalClosures"] = convert._convert_bool(avoid_seasonal_closures)

        if avoid_turns is not None:
            params["avoidTurns"] = avoid_turns

        if allowed_zones is not None:
            params["allowedZones"] = convert._delimit_list(allowed_zones, ',')

        if exclude_zones is not None:
            params["excludeZones"] = convert._delimit_list(exclude_zones, ',')

        if exclude_zone_types is not None:
            params["excludeZoneTypes"] = convert._delimit_list(exclude_zone_types, ',')

        if exclude_countries is not None:
            params["excludeCountries"] = convert._delimit_list(exclude_countries, ',')

        if departure is not None:
            params["departure"] = departure
        elif arrival is not None:
            params["arrival"] = arrival

        if alternatives is not None:
            params["alternatives"] = alternatives

        if metric_system is not None:
            params["metricSystem"] = metric_system

        if view_bounds is not None:
            params["viewBounds"] = convert._delimit_list(
                [
                    convert._delimit_list([convert._format_float(f)
                                           for f in list(reversed(pair))], ',')
                    for pair in view_bounds
                ], ';'
            )

        if resolution is not None:
            params["resolution"] = str(resolution['viewresolution'])
            if 'snapresolution' in resolution:
                params["resolution"] += ':' + str(resolution['snapresolution'])

        if instruction_format is not None:
            params["instructionFormat"] = instruction_format

        if json_attributes is not None:
            params["jsonAttributes"] = json_attributes

        if json_callback is not None:
            params["jsonCallback"] = json_callback

        if representation is not None:
            params["representation"] = convert._delimit_list(representation, ',')

        if route_attributes is not None:
            params["routeAttributes"] = convert._delimit_list(route_attributes, ',')

        if leg_attributes is not None:
            params["legAttributes"] = convert._delimit_list(leg_attributes, ',')

        if maneuver_attributes is not None:
            params["maneuverAttributes"] = convert._delimit_list(maneuver_attributes, ',')

        if link_attributes is not None:
            params["linkAttributes"] = convert._delimit_list(link_attributes, ',')

        if line_attributes is not None:
            params["lineAttributes"] = convert._delimit_list(line_attributes, ',')

        if generalization_tolerances is not None:
            params["generalizationTolerances"] = convert._delimit_list(generalization_tolerances, ',')

        if vehicle_type is not None:
            params["vehicleType"] = vehicle_type

        if license_plate is not None:
            params["licensePlate"] = license_plate

        if max_number_of_changes is not None:
            params["maxNumberOfChanges"] = max_number_of_changes

        if avoid_transport_types is not None:
            params["avoidTransportTypes"] = convert._delimit_list(avoid_transport_types, ',')

        if walk_time_multiplier is not None:
            params["walkTimeMultiplier"] = walk_time_multiplier

        if walk_speed is not None:
            params["walkSpeed"] = walk_speed

        if walk_radius is not None:
            params["walkRadius"] = walk_radius

        if combine_change is not None:
            params["combineChange"] = convert._convert_bool(combine_change)

        if truck_type is not None:
            params["truckType"] = truck_type

        if trailers_count is not None:
            params["trailersCount"] = trailers_count

        if shipped_hazardous_goods is not None:
            params["shippedHazardousGoods"] = convert._delimit_list(shipped_hazardous_goods, ',')

        if limited_weight is not None:
            params["limitedWeight"] = limited_weight

        if weight_per_axle is not None:
            params["weightPerAxle"] = weight_per_axle

        if height is not None:
            params["height"] = height

        if width is not None:
            params["width"] = width

        if length is not None:
            params["length"] = length

        if tunnel_category is not None:
            params["tunnelCategory"] = convert._delimit_list(tunnel_category, ',')

        if truck_restriction_penalty is not None:
            params["truckRestrictionPenalty"] = truck_restriction_penalty

        if return_elevation is not None:
            params["returnElevation"] = convert._convert_bool(return_elevation)

        if consumption_model is not None:
            params["consumptionModel"] = consumption_model

        if custom_consumption_details is not None:
            params["customConsumptionDetails"] = custom_consumption_details

        if speed_profile is not None:
            params["speedProfile"] = speed_profile

        return self._parse_direction_json(
            self._request(
                convert._delimit_list(["/calculateroute", format], '.'),
                get_params=params,
                dry_run=dry_run
            ),
            alternatives=alternatives
        )

    @staticmethod
    def _parse_direction_json(response, alternatives):
        if response is None:  # pragma: no cover
            if alternatives:
                return Directions()
            else:
                return Direction()

        if alternatives is not None and alternatives > 1:
            routes = []
            for route in response['response']['route']:
                routes.append(
                    Direction(
                        geometry=[
                            list(reversed(list((map(float, coordinates.split(','))))))
                            for coordinates in route['shape']
                        ],
                        duration=int(route['summary']['baseTime']),
                        distance=int(route['summary']['distance']),
                        raw=route
                    )
                )

            return Directions(directions=routes, raw=response)

        else:
            geometry = [
                list(reversed(list(map(float, coordinates.split(',')))))
                for coordinates in response['response']['route'][0].get('shape')
            ]
            duration = int(response['response']['route'][0]['summary'].get('baseTime'))
            distance = int(response['response']['route'][0]['summary'].get('distance'))

            return Direction(geometry=geometry, duration=duration, distance=distance, raw=response)

    def isochrones(
            self,
            locations,
            profile,
            intervals,
            mode_type='fastest',
            interval_type='time',
            format='json',
            center_type='start',
            request_id=None,
            arrival=None,
            departure=None,
            single_component=None,
            resolution=None,
            max_points=None,
            quality=None,
            json_attributes=None,
            json_callback=None,
            truck_type=None,
            trailers_count=None,
            shipped_hazardous_goods=None,
            limited_weight=None,
            weight_per_axle=None,
            height=None,
            width=None,
            length=None,
            tunnel_category=None,
            consumption_model=None,
            custom_consumption_details=None,
            speed_profile=None,
            dry_run=None
    ):
        """Gets isochrones or equidistants for a range of time/distance values around a given set of coordinates.

        For more information, https://developer.here.com/documentation/routing/topics/resource-calculate-isoline.html.

        :param locations: One pair of lng/lat values.
        :type locations: list of float

        :param profile: Specifies the routing mode of transport and further options.
            Can be a str or :class:`HereMaps.RoutingMode`
            https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html
        :type profile: str or :class:`HereMaps.RoutingMode`

        :param intervals: Range of isoline. Several comma separated values can be specified. 
            The unit is defined by parameter rangetype.
        :type ranges: list of int

        :param mode_type: RoutingType relevant to calculation. One of [fastest, shortest, balanced]. Default fastest.
            https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html#ariaid-title2
        :type mode_type: str

        :param interval_type: Specifies type of range. Possible values are distance, 
            time, consumption. For distance the unit is meters. For time the unit is seconds. 
            For consumption it is defined by consumption model
        :type range_type: str

        :param format: Currently only "json" supported. 
        :type format: str

        :param center_type: If 'start' then the isoline will cover all roads which 
            can be reached from this point within given range.
            It cannot be used in combination with destination parameter. 
            If 'destination' Center of the isoline request. Isoline will cover all 
            roads from which this point can be reached within given range. 
            It cannot be used in combination with start parameter. 
        :type center_type: str

        :param departure: Time when travel is expected to start. Traffic speed and 
            incidents are taken into account
            when calculating the route (note that in case of a past
            departure time the historical traffic is limited to one year).  
            You can use now to specify the current time. Specify either departure 
            or arrival, not both. When the optional timezone offset is not 
            specified, the time is assumed to be the local.
            Formatted as iso time, e.g. 2018-07-04T17:00:00+02.
        :type departure: str

        :param arrival: Time when travel is expected to end. Specify either 
            departure or arrival, not both.
            When the optional timezone offset is not specified, the time is 
            assumed to be the local.
            Formatted as iso time, e.g. 2018-07-04T17:00:00+02.
        :type arrival: str        

        :param single_component: If set to true the isoline service will always 
            return single polygon, instead of creating a separate polygon 
            for each ferry separated island. Default value is false.
        :type single_component: bool

        :param resolution: Allows to specify level of detail needed for the
            isoline polygon. Unit is meters per pixel. Higher resolution may 
            cause increased response time from the service.
        :type resolution: int

        :param max_points: Allows to limit amount of points in the returned 
            isoline. If isoline consists of multiple components, sum of points from 
            all components is considered. Each component will have at least 2 points, 
            so it is possible that more points than maxpoints value will be returned. 
            This is in case when 2 * number of components is higher than maxpoints. 
            Enlarging number of maxpoints may cause increased response time from the service.
        :type max_points: int

        :param quality: Allows to reduce the quality of the isoline in favor 
            of the response time. Allowed values are 1, 2, 3. 
            Default value is 1 and it is the best quality.
        :type quality: int

        :param json_attributes: Flag to control JSON output. 
            Combine parameters by adding their values. 
            https://developer.here.com/documentation/routing/topics/resource-param-type-json-representation.html
        :type json_attributes: int

        :param truck_type: Truck routing only, specifies the vehicle type. Defaults to truck.
        :type truck_type: str

        :param trailers_count: Truck routing only, specifies number of trailers 
            pulled by a vehicle. The provided value must be between 0 and 4. Defaults to 0.
        :type trailers_count: int

        :param shipped_hazardous_goods: Truck routing only, list of hazardous 
            materials in the vehicle. Please refer to the enumeration type 
            HazardousGoodTypeType for available values. 
            Note the value allhazardousGoods does not apply to the request parameter.
            https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html#resource-type-enumerations__enum-hazardous-good-type-type
        :type shipped_hazardous_goods: list of str

        :param limited_weight: Truck routing only, vehicle weight including 
            trailers and shipped goods, in tons. The provided value must be between 0 and 1000.
        :type limited_weight: int

        :param weight_per_axle: Truck routing only, vehicle weight per axle in tons. 
            The provided value must be between 0 and 1000.
        :type limited_weight: int

        :param height: Truck routing only, vehicle height in meters. 
            The provided value must be between 0 and 50.
        :type height: int

        :param width: Truck routing only, vehicle width in meters. 
            The provided value must be between 0 and 50.
        :type width: int  

        :param length: Truck routing only, vehicle length in meters. 
            The provided value must be between 0 and 300.
        :type length: int        

        :param tunnel_category: Truck routing only, specifies the tunnel category to 
            restrict certain route links. The route will pass only through tunnels of a less strict category.
        :type tunnel_category: list of str

        :param consumption_model: If you request information on consumption, you must 
            provide a consumption model. The possible values are default and standard. 
            When you specify the value standard, you must provide additional 
            information in the query parameter customconsumptiondetails
        :type consumption_model: str

        :param custom_consumption_details: Provides vehicle specific information 
            for use in the consumption model. This information can include such 
            things as the amount of energy consumed while travelling at a given speed.
            https://developer.here.com/documentation/routing/topics/resource-param-type-custom-consumption-details.html#type-standard
        :type custom_consumption_details: str

        :param speed_profile: Specifies the speed profile variant for a given 
            routing mode. The speed profile affects travel time estimation as 
            well as roads evaluation when computing the fastest route. 
            Note that computed routes might differ depending on a used profile.
            https://developer.here.com/documentation/routing/topics/resource-param-type-speed-profile-type.html
        :type speed_profile: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: raw JSON response
        :rtype: dict
        """

        self.base_url = 'https://isoline.route.api.here.com/routing/7.2'
        params = {}

        params["app_code"] = self.app_code
        params["app_id"] = self.app_id

        params[center_type] = self._build_locations(locations)[0]

        if isinstance(profile, str):
            params["mode"] = mode_type + ';' + profile
        elif isinstance(profile, self.RoutingMode):
            params["mode"] = profile.make_routing_mode()

        if intervals is not None:
            params["range"] = convert._delimit_list(intervals, ',')

        if interval_type is not None:
            params["rangeType"] = interval_type

        if departure is not None:
            params["departure"] = departure
        elif arrival is not None:
            params["arrival"] = arrival

        if single_component is not None:
            params["singleComponent"] = convert._convert_bool(single_component)

        if max_points is not None:
            params["maxPoints"] = max_points

        if quality is not None:
            params["quality"] = quality

        if json_attributes is not None:
            params["jsonAttributes"] = json_attributes

        if json_callback is not None:
            params["jsonCallback"] = json_callback

        if truck_type is not None:
            params["truckType"] = truck_type

        if trailers_count is not None:
            params["trailersCount"] = trailers_count

        if shipped_hazardous_goods is not None:
            params["shippedHazardousGoods"] = convert._delimit_list(shipped_hazardous_goods, ',')

        if limited_weight is not None:
            params["limitedWeight"] = limited_weight

        if weight_per_axle is not None:
            params["weightPerAxle"] = weight_per_axle

        if height is not None:
            params["height"] = height

        if width is not None:
            params["width"] = width

        if length is not None:
            params["length"] = length

        if tunnel_category is not None:
            params["tunnelCategory"] = convert._delimit_list(tunnel_category, ',')

        if consumption_model is not None:
            params["consumptionModel"] = consumption_model

        if custom_consumption_details is not None:
            params["customConsumptionDetails"] = custom_consumption_details

        if speed_profile is not None:
            params["speedProfile"] = speed_profile

        return self._parse_isochrone_json(
            self._request(
                convert._delimit_list(["/calculateisoline", format], '.'),
                get_params=params,
                dry_run=dry_run
            ), intervals
        )

    @staticmethod
    def _parse_isochrone_json(response, intervals):
        if response is None:  # pragma: no cover
            return Isochrones()

        geometries = []
        for idx, isochrones in enumerate(response['response']['isoline']):
            range_polygons = []
            if 'component' in isochrones:
                for component in isochrones['component']:
                    if 'shape' in component:
                        coordinates_list = []
                        for coordinates in component['shape']:
                            coordinates_list.append([float(f) for f in coordinates.split(",")])
                        range_polygons.append(coordinates_list)

            geometries.append(
                Isochrone(
                    geometry=range_polygons,
                    interval=intervals[idx],
                    center=list(response['response']['start']['mappedPosition'].values())
                )
            )

        return Isochrones(isochrones=geometries, raw=response)

    def matrix(
            self,
            locations,
            profile,
            format='json',
            mode_type='fastest',
            sources=None,
            destinations=None,
            search_range=None,
            avoid_areas=None,
            avoid_links=None,
            avoid_turns=None,
            exclude_countries=None,
            departure=None,
            matrix_attributes=None,
            summary_attributes=['traveltime', 'costfactor', 'distance'],
            truck_type=None,
            trailers_count=None,
            shipped_hazardous_goods=None,
            limited_weight=None,
            weight_per_axle=None,
            height=None,
            width=None,
            length=None,
            tunnel_category=None,
            speed_profile=None,
            dry_run=None
    ):
        """ Gets travel distance and time for a matrix of origins and destinations.

            :param locations: The coordinates tuple the route should be calculated
                from in order of visit. Can be a list/tuple of [lon, lat] or :class:`HereMaps.Waypoint` instance or a
                combination of those. For further explanation, see
                https://developer.here.com/documentation/routing/topics/resource-param-type-waypoint.html
            :type locations: list of list or list of :class:`HereMaps.Waypoint`

            :param profile: Specifies the routing mode of transport and further options.
                Can be a str or :class:`HereMaps.RoutingMode`
                https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html
            :type profile: str or :class:`HereMaps.RoutingMode`

            :param mode_type: RoutingType relevant to calculation. One of [fastest, shortest, balanced]. Default fastest.
                https://developer.here.com/documentation/routing/topics/resource-param-type-routing-mode.html#ariaid-title2
            :type mode_type: str

            :param sources: The starting points for the matrix. 
                Specifies an index referring to coordinates.
            :type sources: list of int

            :param destinations: The destination points for the routes. 
                Specifies an index referring to coordinates.
            :type destinations: list of int

            :param search_range: Defines the maximum search range for destination 
                waypoints, in meters. This parameter is especially useful for optimizing 
                matrix calculation where the maximum desired effective distance is known 
                in advance. Destination waypoints with a longer effective distance than 
                specified by searchRange will be skipped. The parameter is optional. 
                In pedestrian mode the default search range is 20 km. 
                If parameter is omitted in other modes, no range limit will apply.
            :type search_range: int

            :param avoid_areas: Areas which the route must not cross. 
                Array of BoundingBox. Example with 2 bounding boxes
                https://developer.here.com/documentation/routing/topics/resource-param-type-bounding-box.html
            :type avoid_areas: list of list of list

            :param avoid_links: Links which the route must not cross. 
              The list of LinkIdTypes.
            :type avoid_areas: list of string

            :param avoid_turns: List of turn types that the route should avoid. Defaults to empty list. 
              https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html
            :type avoid_turns: str  

            :param exclude_countries: Countries that must be excluded from route calculation. 
            :type exclude_countries: list of str

            :param departure: Time when travel is expected to start. Traffic speed and 
                incidents are taken into account
                when calculating the route (note that in case of a past
                departure time the historical traffic is limited to one year).  
                You can use now to specify the current time. Specify either departure 
                or arrival, not both. When the optional timezone offset is not 
                specified, the time is assumed to be the local.
                Formatted as iso time, e.g. 2018-07-04T17:00:00+02.
            :type departure: str

            :param matrix_attributes: Defines which attributes are included in the 
              response as part of the data representation of the route matrix entries. 
              Defaults to indices and summary. 
              https://developer.here.com/documentation/routing/topics/resource-calculate-matrix.html#resource-calculate-matrix__matrix-route-attribute-type
            :type matrix_attributes: list of str

            :param summary_attributes: Defines which attributes are included in 
                the response as part of the data representation of the matrix 
                entries summaries. Defaults to costfactor.  
                https://developer.here.com/documentation/routing/topics/resource-calculate-matrix.html#resource-calculate-matrix__matrix-route-summary-attribute-type
            :type matrix_attributes: list of str

            :param truck_type: Truck routing only, specifies the vehicle type. 
                Defaults to truck.
            :type truck_type: str

            :param trailers_count: Truck routing only, specifies number of 
                trailers pulled by a vehicle. The provided value must be between 0 and 4. 
                Defaults to 0.
            :type trailers_count: int

            :param shipped_hazardous_goods: Truck routing only, list of hazardous
                materials in the vehicle. Please refer to the enumeration type 
                HazardousGoodTypeType for available values. Note the value 
                allhazardousGoods does not apply to the request parameter.
                https://developer.here.com/documentation/routing/topics/resource-type-enumerations.html#resource-type-enumerations__enum-hazardous-good-type-type
            :type shipped_hazardous_goods: list of str

            :param limited_weight: Truck routing only, vehicle weight including 
                trailers and shipped goods, in tons. The provided value must be 
                between 0 and 1000.
            :type limited_weight: int

            :param weight_per_axle: Truck routing only, vehicle weight per axle 
                in tons. The provided value must be between 0 and 1000.
            :type limited_weight: int

            :param height: Truck routing only, vehicle height in meters. The 
                provided value must be between 0 and 50.
            :type height: int

            :param width: Truck routing only, vehicle width in meters. 
                The provided value must be between 0 and 50.
            :type width: int  

            :param length: Truck routing only, vehicle length in meters. 
                The provided value must be between 0 and 300.
            :type length: int        

            :param tunnel_category: Truck routing only, specifies the tunnel 
                category to restrict certain route links. The route will pass 
                only through tunnels of a less strict category.
            :type tunnel_category: list of str

            :param speed_profile: Specifies the speed profile variant for a given 
                routing mode. The speed profile affects travel time estimation as 
                well as roads evaluation when computing the fastest route. 
                Note that computed routes might differ depending on a used profile.
                https://developer.here.com/documentation/routing/topics/resource-param-type-speed-profile-type.html
            :type speed_profile: str

            :param dry_run: Print URL and parameters without sending the request.
            :param dry_run: bool

            :returns: raw JSON response
            :rtype: dict
            """
        self.base_url = 'https://matrix.route.api.here.com/routing/7.2'
        params = {}

        params["app_code"] = self.app_code
        params["app_id"] = self.app_id

        locations = self._build_locations(locations)

        sources_coords = locations
        if sources is not None:
            sources_coords = itemgetter(*sources)(sources_coords)
            if isinstance(sources_coords, str):
                sources_coords = [sources_coords]
        for i, location in enumerate(sources_coords):
            params["start" + str(i)] = location

        dest_coords = locations
        if destinations is not None:
            dest_coords = itemgetter(*destinations)(dest_coords)
            if isinstance(dest_coords, str):
                dest_coords = [dest_coords]
        for i, location in enumerate(dest_coords):
            params["destination" + str(i)] = location

        if isinstance(profile, str):
            params["mode"] = mode_type + ';' + profile
        elif isinstance(profile, self.RoutingMode):
            params["mode"] = profile.make_routing_mode()

        if search_range is not None:
            params["searchRange"] = search_range

        if avoid_areas is not None:
            params["avoidAreas"] = convert._delimit_list(
                [
                    convert._delimit_list(
                        [
                            convert._delimit_list(
                                [convert._format_float(f) for f in list(reversed(pair))], ','
                            ) for pair in bounding_box
                        ], ';'
                    ) for bounding_box in avoid_areas
                ], '!'
            )

        if avoid_links is not None:
            params["avoidLinks"] = convert._delimit_list(avoid_links, ',')

        if avoid_turns is not None:
            params["avoidTurns"] = avoid_turns

        if exclude_countries is not None:
            params["excludeCountries"] = convert._delimit_list(exclude_countries, ',')

        if departure is not None:
            params["departure"] = departure.isoformat()

        if matrix_attributes is not None:
            params["matrixAttributes"] = convert._delimit_list(matrix_attributes, ',')

        if summary_attributes is not None:
            params["summaryAttributes"] = convert._delimit_list(summary_attributes, ',')

        if truck_type is not None:
            params["truckType"] = truck_type

        if trailers_count is not None:
            params["trailersCount"] = trailers_count

        if shipped_hazardous_goods is not None:
            params["shippedHazardousGoods"] = convert._delimit_list(shipped_hazardous_goods, ',')

        if limited_weight is not None:
            params["limitedWeight"] = limited_weight

        if weight_per_axle is not None:
            params["weightPerAxle"] = weight_per_axle

        if height is not None:
            params["height"] = height

        if width is not None:
            params["width"] = width

        if length is not None:
            params["length"] = length

        if tunnel_category is not None:
            params["tunnelCategory"] = convert._delimit_list(tunnel_category, ',')

        if speed_profile is not None:
            params["speedProfile"] = speed_profile

        return self._parse_matrix_json(
            self._request(
                convert._delimit_list(["/calculatematrix", format], '.'),
                get_params=params,
                dry_run=dry_run
            )
        )

    @staticmethod
    def _parse_matrix_json(response):
        if response is None:  # pragma: no cover
            return Matrix()

        durations = []
        distances = []
        index_durations = []
        index_distances = []

        next_ = None
        mtx_objects = response['response']['matrixEntry']
        l = len(mtx_objects)
        for index, obj in enumerate(mtx_objects):
            if index < (l - 1):
                next_ = mtx_objects[index + 1]

            if 'travelTime' in obj['summary']:
                index_durations.append(obj['summary']['travelTime'])
            else:
                index_durations.append(obj['summary']['costfactor'])

            if 'distance' in obj['summary']:
                index_distances.append(obj['summary']['distance'])

            if next_['startIndex'] > obj['startIndex']:
                durations.append(index_durations)
                distances.append(index_distances)
                index_durations = []
                index_distances = []

        durations.append(index_durations)
        distances.append(index_distances)

        return Matrix(durations=durations, distances=distances, raw=response)

    def _build_locations(self, coordinates, matrix=False):
        """Build the locations object for all methods"""

        locations = []

        # Directions and matrix calls which are lists of list
        if isinstance(coordinates[0], (list, tuple, self.Waypoint)):

            for idx, coord in enumerate(coordinates):
                if isinstance(locations, self.Waypoint):
                    locations.append(locations._make_waypoint())
                elif isinstance(locations, (list, tuple)):
                    wp = 'geo!' + convert._delimit_list(
                        [convert._format_float(f) for f in list(reversed(coord))], ','
                    )
                    locations.append(wp)
                else:
                    raise TypeError(
                        "Location type {} at index {} is not supported: {}".format(
                            type(coord), idx, coord
                        )
                    )

        # Isochrones
        elif isinstance(coordinates[0], float):
            center = 'geo!' + convert._delimit_list(
                [convert._format_float(f) for f in list(reversed(coordinates))], ','
            )
            locations.append(center)
        # Isochrones using waypoint class
        elif isinstance(coordinates, self.Waypoint):
            locations.append(coordinates._make_waypoint())

        return locations



"""
Every routing service below has a separate module in ``routingpy.routers``, which hosts a class which
bases :class:`routingpy.routers.base.Router` astracting the service's API. Each router
has at least a ``directions`` method, many offer additionally ``matrix`` and/or ``isochrones`` methods. Other
available provider endpoints are allowed and generally encouraged. However, please refer
to our `contribution guidelines`_ for general instructions.

**routingpy**'s dogma is, that all routers expose the same mandatory arguments for common methods in an
attempt to be consistent for the same method across different routers. Unlike other collective libraries,
we additionally chose to preserve each router's special arguments, only abstracting the most basic arguments, such as
locations and profile (car, bike, pedestrian etc.), among others (full list here_).

.. _`contribution_guidelines`: https://github.com/gis-ops/routing-py/blob/master/CONTRIBUTING.md
.. _here: https://github.com/gis-ops/routing-py#api
"""
from .base import options
from routingpy.exceptions import RouterNotFound

from .openrouteservice import ORS
from .osrm import OSRM
from .valhalla import Valhalla
from .graphhopper import Graphhopper
from .mapbox_valhalla import MapboxValhalla
from .mapbox_osrm import MapboxOSRM
from .google import Google
from .heremaps import HereMaps

# Provide synonyms
_SERVICE_TO_ROUTER = {
    "ors": ORS,
    "openrouteservice": ORS,
    "osrm": OSRM,
    "mapbox_osrm": MapboxOSRM,
    "mapbox-osrm": MapboxOSRM,
    "mapboxosrm": MapboxOSRM,
    "mapbox": MapboxOSRM,
    "valhalla": Valhalla,
    "mapbox_valhalla": MapboxValhalla,
    "mapbox-valhalla": MapboxValhalla,
    "mapboxvalhalla": MapboxValhalla,
    "graphhopper": Graphhopper,
    "google": Google,
    "here": HereMaps,
    "heremaps": HereMaps
}


def get_router_by_name(router_name):
    """
    Given a router's name, try to return the router class.

    >>> from routingpy.routers import get_router_by_name
    >>> router = get_router_by_name("ors")(api_key='')
    >>> print(router)
    routingpy.routers.openrouteservice.ORS
    >>> route = router.directions(**params)

    If the string given is not recognized, a
    :class:`routingpy.exceptions.RouterNotFound` exception is raised and the available list of router names is printed.

    :param router_name: Name of the router as string.
    :type router_name: str

    :rtype: :class:`routingpy.routers.base.Router`

    """
    try:
        return _SERVICE_TO_ROUTER[router_name.lower()]
    except KeyError:
        raise RouterNotFound(
            "Unknown router '{}'; options are: {}".format(router_name, _SERVICE_TO_ROUTER.keys())
        )



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

from .base import Router, DEFAULT
from routingpy import convert, utils
from routingpy.direction import Directions, Direction
from routingpy.matrix import Matrix

from operator import itemgetter


class Google(Router):
    """Performs requests to the Google API services."""

    _base_url = "https://maps.googleapis.com/maps/api"

    def __init__(
            self,
            api_key,
            user_agent=None,
            timeout=DEFAULT,
            retry_timeout=None,
            requests_kwargs={},
            retry_over_query_limit=True,
            skip_api_error=None
    ):
        """
        Initializes a Google client.

        :param key: API key.
        :type key: str

        :param user_agent: User Agent to be used when requesting.
            Default :attr:`routingpy.routers.options.default_user_agent`.
        :type user_agent: str

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify ``None`` for no timeout. Default :attr:`routingpy.routers.options.default_timeout`.
        :type timeout: int or None

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.  Default :attr:`routingpy.routers.options.default_retry_timeout`.
        :type retry_timeout: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. **Note**, that ``proxies`` can be set globally
            in :attr:`routingpy.routers.options.default_proxies`.

            Example:

            >>> from routingpy.routers import Google
            >>> router = Google(my_key, requests_kwargs={'proxies': {'https': '129.125.12.0'}})
            >>> print(router.proxies)
            {'https': '129.125.12.0'}
        :type requests_kwargs: dict

        :param retry_over_query_limit: If True, client will not raise an exception
            on HTTP 429, but instead jitter a sleeping timer to pause between
            requests until HTTP 200 or retry_timeout is reached.
            Default :attr:`routingpy.routers.options.default_over_query_limit`.
        :type retry_over_query_limit: bool

        :param skip_api_error: Continue with batch processing if a :class:`routingpy.exceptions.RouterApiError` is
            encountered (e.g. no route found). If False, processing will discontinue and raise an error.
            Default :attr:`routingpy.routers.options.default_skip_api_error`.
        :type skip_api_error: bool
        """

        self.key = api_key

        super(Google, self).__init__(
            self._base_url, user_agent, timeout, retry_timeout, requests_kwargs, retry_over_query_limit,
            skip_api_error
        )

    class WayPoint(object):
        """
        TODO: make the WayPoint class and its parameters appear in Sphinx. True for Valhalla as well.

        Optionally construct a waypoint from this class with additional attributes.

        Example:

        >>> waypoint = Google.WayPoint(position=[8.15315, 52.53151], waypoint_type='coords', stopover=False)
        >>> route = Google(api_key).directions(locations=[[[8.58232, 51.57234]], waypoint, [7.15315, 53.632415]])
        """

        def __init__(self, position, waypoint_type='coords', stopover=True):
            """
            Constructs a waypoint with additional information, such as via or encoded lines.

            :param position: Coordinates in [long, lat] order.
            :type position: list/tuple of float

            :param waypoint_type: The type of information provided. One of ['place_id', 'enc', 'coords']. Default 'coords'.
            :type waypoint_type: str

            :param stopover: If True, the waypoint will be used to add an additional leg to the journey. If False,
                it's only used as a via waypoint. Not supported for first and last waypoint. Default True.
            :type stopover: bool
            """

            self.position = position
            self.waypoint_type = waypoint_type
            self.stopover = stopover

        def make_waypoint(self):

            waypoint = ''
            if self.waypoint_type == 'coords':
                waypoint += convert._delimit_list(list(reversed(self.position)))
            elif self.waypoint_type == 'place_id':
                waypoint += self.waypoint_type + ':' + self.position
            elif self.waypoint_type == 'enc':
                waypoint += self.waypoint_type + ':' + self.position + ':'
            else:
                raise ValueError("waypoint_type only supports enc, place_id, coords")

            if not self.stopover:
                waypoint = 'via:' + waypoint

            return waypoint

    def directions(
            self,
            locations,
            profile,
            alternatives=None,
            avoid=None,
            optimize=None,
            language=None,
            region=None,
            units=None,
            arrival_time=None,
            departure_time=None,
            traffic_model=None,
            transit_mode=None,
            transit_routing_preference=None,
            dry_run=None
    ):
        """Get directions between an origin point and a destination point.

        For more information, visit https://developers.google.com/maps/documentation/directions/intro.

        :param locations: The coordinates tuple the route should be calculated
            from in order of visit. Can be a list/tuple of [lon, lat], a list/tuple of address strings, Google's
            Place ID's, a :class:`Google.WayPoint` instance or a combination of these. Note, the first and last location have to be specified as [lon, lat].
            Optionally, specify ``optimize=true`` for via waypoint optimization.
        :type locations: list of list or list of :class:`Google.WayPoint`

        :param profile: The vehicle for which the route should be calculated.
            Default "driving". One of ['driving', 'walking', 'bicycling', 'transit'].
        :type profile: str

        :param alternatives: Specifies whether more than one route should be returned.
            Only available for requests without intermediate waypoints. Default False.
        :type alternatives: bool

        :param avoid: Indicates that the calculated route(s) should avoid the indicated features. One or more of
            ['tolls', 'highways', 'ferries', 'indoor']. Default None.
        :param avoid: list of str

        :param optimize: Optimize the given order of via waypoints (i.e. between first and last location). Default False.
        :type optimize: bool

        :param language: Language for routing instructions. The locale of the resulting turn instructions. Visit
            https://developers.google.com/maps/faq#languagesupport for options.
        :type language: str

        :param region: Specifies the region code, specified as a ccTLD ("top-level domain") two-character value.
            See https://developers.google.com/maps/documentation/directions/intro#RegionBiasing.
        :type region: str

        :param units: Specifies the unit system to use when displaying results. One of ['metric', 'imperial'].
        :type units: str

        :param arrival_time: Specifies the desired time of arrival for transit directions, in seconds since midnight,
            January 1, 1970 UTC. Incompatible with departure_time.
        :type arrival_time: int

        :param departure_time: Specifies the desired time of departure. You can specify the time as an integer in
            seconds since midnight, January 1, 1970 UTC.

        :param traffic_model: Specifies the assumptions to use when calculating time in traffic. One of ['best_guess',
            'pessimistic', 'optimistic'. See https://developers.google.com/maps/documentation/directions/intro#optional-parameters
            for details.
        :type traffic_model: str

        :param transit_mode: Specifies one or more preferred modes of transit. One or more of ['bus', 'subway', 'train',
            'tram', 'rail'].
        :type transit_mode: list/tuple of str

        :param transit_routing_preference: Specifies preferences for transit routes. Using this parameter, you can bias
            the options returned, rather than accepting the default best route chosen by the API. One of ['less_walking',
            'fewer_transfers'].
        :type transit_routing_preference: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: One or multiple route(s) from provided coordinates and restrictions.
        :rtype: :class:`routingpy.direction.Direction` or :class:`routingpy.direction.Directions`
        """

        params = {'profile': profile}

        origin, destination = locations[0], locations[-1]
        if isinstance(origin, (list, tuple)):
            params['origin'] = convert._delimit_list(list(reversed(origin)))
        elif isinstance(origin, self.WayPoint):
            raise TypeError("The first and last locations must be list/tuple of [lon, lat]")

        if isinstance(destination, (list, tuple)):
            params['destination'] = convert._delimit_list(list(reversed(destination)))
        elif isinstance(origin, self.WayPoint):
            raise TypeError("The first and last locations must be list/tuple of [lon, lat]")

        if len(locations) > 2:
            waypoints = []
            s = slice(1, -1)
            for coord in locations[s]:
                if isinstance(coord, (list, tuple)):
                    waypoints.append(convert._delimit_list(list(reversed(coord))))
                elif isinstance(coord, self.WayPoint):
                    waypoints.append(coord.make_waypoint())
            if optimize:
                waypoints.insert(0, 'optimize:true')

            params['waypoints'] = convert._delimit_list(waypoints, '|')

        if self.key is not None:
            params["key"] = self.key

        if alternatives is not None:
            params['alternatives'] = convert._convert_bool(alternatives)

        if avoid:
            params['avoid'] = convert._delimit_list(avoid, '|')

        if language:
            params['language'] = language

        if region:
            params['region'] = region

        if units:
            params['units'] = units

        if arrival_time and departure_time:
            raise ValueError("Specify either arrival_time or departure_time.")

        if arrival_time:
            params['arrival_time'] = str(arrival_time)

        if departure_time:
            params['departure_time'] = str(departure_time)

        if traffic_model:
            params['traffic_model'] = traffic_model

        if transit_mode:
            params['transit_mode'] = convert._delimit_list(transit_mode, '|')

        if transit_routing_preference:
            params['transit_routing_preference'] = transit_routing_preference

        return self._parse_direction_json(
            self._request('/directions/json', get_params=params, dry_run=dry_run), alternatives
        )

    @staticmethod
    def _parse_direction_json(response, alternatives):
        if response is None:  # pragma: no cover
            if alternatives:
                return Directions()
            else:
                return Direction()

        if alternatives:
            routes = []
            for route in response['routes']:
                geometry = []
                duration, distance = 0, 0
                for leg in route['legs']:
                    duration += leg['duration']['value']
                    distance += leg['distance']['value']
                    for step in leg['steps']:
                        geometry.extend(
                            [
                                list(reversed(coords))
                                for coords in utils.decode_polyline5(step['polyline']['points'])
                            ]
                        )
                routes.append(
                    Direction(
                        geometry=geometry, duration=int(duration), distance=int(distance), raw=route
                    )
                )
            return Directions(routes, response)
        else:
            geometry = []
            duration, distance = 0, 0
            for leg in response['routes'][0]['legs']:
                duration = int(leg['duration']['value'])
                distance = int(leg['distance']['value'])
                for step in leg['steps']:
                    geometry.extend(
                        [
                            list(reversed(coords))
                            for coords in utils.decode_polyline5(step['polyline']['points'])
                        ]
                    )
            return Direction(geometry=geometry, duration=duration, distance=distance, raw=response)

    def isochrones(self):  # pragma: no cover
        raise NotImplementedError

    def matrix(
            self,
            locations,
            profile,
            sources=None,
            destinations=None,
            avoid=None,
            language=None,
            region=None,
            units=None,
            arrival_time=None,
            departure_time=None,
            traffic_model=None,
            transit_mode=None,
            transit_routing_preference=None,
            dry_run=None
    ):
        """ Gets travel distance and time for a matrix of origins and destinations.

        :param locations: Two or more pairs of lng/lat values.
        :type locations: list of list

        :param profile: The vehicle for which the route should be calculated.
            Default "driving". One of ['driving', 'walking', 'bicycling', 'transit'].
        :type profile: str

        :param sources: A list of indices that refer to the list of locations
            (starting with 0). If not passed, all indices are considered.
        :type sources: list or tuple

        :param destinations: A list of indices that refer to the list of locations
            (starting with 0). If not passed, all indices are considered.
        :type destinations: list or tuple

        :param avoid: Indicates that the calculated route(s) should avoid the indicated features. One or more of
            ['tolls', 'highways', 'ferries', 'indoor']. Default None.
        :param avoid: list of str

        :param language: Language for routing instructions. The locale of the resulting turn instructions. Visit
            https://developers.google.com/maps/faq#languagesupport for options.
        :type language: str

        :param region: Specifies the region code, specified as a ccTLD ("top-level domain") two-character value.
            See https://developers.google.com/maps/documentation/directions/intro#RegionBiasing.
        :type region: str

        :param units: Specifies the unit system to use when displaying results. One of ['metric', 'imperial'].
        :type units: str

        :param arrival_time: Specifies the desired time of arrival for transit directions, in seconds since midnight,
            January 1, 1970 UTC. Incompatible with departure_time.
        :type arrival_time: int

        :param departure_time: Specifies the desired time of departure. You can specify the time as an integer in
            seconds since midnight, January 1, 1970 UTC.
        :type departure_time: int

        :param traffic_model: Specifies the assumptions to use when calculating time in traffic. One of ['best_guess',
            'pessimistic', 'optimistic'. See https://developers.google.com/maps/documentation/directions/intro#optional-parameters
            for details.
        :type traffic_model: str

        :param transit_mode: Specifies one or more preferred modes of transit. One or more of ['bus', 'subway', 'train',
            'tram', 'rail'].
        :type transit_mode: list of str or tuple of str

        :param transit_routing_preference: Specifies preferences for transit routes. Using this parameter, you can bias
            the options returned, rather than accepting the default best route chosen by the API. One of ['less_walking',
            'fewer_transfers'].
        :type transit_routing_preference: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: A matrix from the specified sources and destinations.
        :rtype: :class:`routingpy.matrix.Matrix`
        """
        params = {'profile': profile}

        waypoints = []
        for coord in locations:
            if isinstance(coord, (list, tuple)):
                waypoints.append(convert._delimit_list(list(reversed(coord))))
            elif isinstance(coord, self.WayPoint):
                waypoints.append(coord.make_waypoint())

        sources_coords = waypoints
        if sources is not None:
            sources_coords = itemgetter(*sources)(sources_coords)
            if not isinstance(sources_coords, (list, tuple)):
                sources_coords = [sources_coords]
        params['origins'] = convert._delimit_list(sources_coords, '|')

        destinations_coords = waypoints
        if destinations is not None:
            destinations_coords = itemgetter(*destinations)(destinations_coords)
            if not isinstance(destinations_coords, (list, tuple)):
                destinations_coords = [destinations_coords]
        params['destinations'] = convert._delimit_list(destinations_coords, '|')

        if self.key is not None:
            params["key"] = self.key

        if avoid:
            params['avoid'] = convert._delimit_list(avoid, '|')

        if language:
            params['language'] = language

        if region:
            params['region'] = region

        if units:
            params['units'] = units

        if arrival_time:
            params['arrival_time'] = str(arrival_time)

        if departure_time:
            params['departure_time'] = str(departure_time)

        if traffic_model:
            params['traffic_model'] = traffic_model

        if transit_mode:
            params['transit_mode'] = convert._delimit_list(transit_mode, '|')

        if transit_routing_preference:
            params['transit_routing_preference'] = transit_routing_preference

        return self._parse_matrix_json(
            self._request('/distancematrix/json', get_params=params, dry_run=dry_run)
        )

    @staticmethod
    def _parse_matrix_json(response):
        if response is None:  # pragma: no cover
            return Matrix()

        durations = [
            [destination['duration']['value']
             for destination in origin['elements']]
            for origin in response['rows']
        ]
        distances = [
            [destination['distance']['value']
             for destination in origin['elements']]
            for origin in response['rows']
        ]

        return Matrix(durations, distances, response)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

from .base import Router, DEFAULT
from routingpy import utils
from routingpy.direction import Direction, Directions
from routingpy.isochrone import Isochrone, Isochrones
from routingpy.matrix import Matrix


class ORS(Router):
    """Performs requests to the ORS API services."""

    _DEFAULT_BASE_URL = 'https://api.openrouteservice.org'

    def __init__(
            self,
            api_key=None,
            base_url=_DEFAULT_BASE_URL,
            user_agent=None,
            timeout=DEFAULT,
            retry_timeout=None,
            requests_kwargs=None,
            retry_over_query_limit=False,
            skip_api_error=None
    ):
        """
        Initializes an openrouteservice client.

        :param api_key: ORS API key. Required if https://api.openrouteservice.org is used.
        :type api_key: str

        :param base_url: The base URL for the request. Defaults to the ORS API
            server. Should not have a trailing slash.
        :type base_url: str

        :param user_agent: User Agent to be used when requesting.
            Default :attr:`routingpy.routers.options.default_user_agent`.
        :type user_agent: str

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify ``None`` for no timeout. Default :attr:`routingpy.routers.options.default_timeout`.
        :type timeout: int or None

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.  Default :attr:`routingpy.routers.options.default_retry_timeout`.
        :type retry_timeout: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. **Note**, that ``proxies`` can be set globally
            in :attr:`routingpy.routers.options.default_proxies`.

            Example:

            >>> from routingpy.routers import ORS
            >>> router = ORS(my_key, requests_kwargs={'proxies': {'https': '129.125.12.0'}})
            >>> print(router.proxies)
            {'https': '129.125.12.0'}
        :type requests_kwargs: dict

        :param retry_over_query_limit: If True, client will not raise an exception
            on HTTP 429, but instead jitter a sleeping timer to pause between
            requests until HTTP 200 or retry_timeout is reached.
            Default :attr:`routingpy.routers.options.default_over_query_limit`.
        :type retry_over_query_limit: bool

        :param skip_api_error: Continue with batch processing if a :class:`routingpy.exceptions.RouterApiError` is
            encountered (e.g. no route found). If False, processing will discontinue and raise an error.
            Default :attr:`routingpy.routers.options.default_skip_api_error`.
        :type skip_api_error: bool
        """

        if base_url == self._DEFAULT_BASE_URL and api_key is None:
            raise KeyError("API key must be specified.")

        requests_kwargs = requests_kwargs or {}
        headers = requests_kwargs.get('headers') or {}
        headers.update({'Authorization': api_key})
        requests_kwargs.update({'headers': headers})

        super(ORS, self).__init__(
            base_url, user_agent, timeout, retry_timeout, requests_kwargs, retry_over_query_limit,
            skip_api_error
        )

    def directions(
            self,
            locations,
            profile,
            format='geojson',
            preference=None,
            alternative_routes=None,
            units=None,
            language=None,
            geometry=None,
            geometry_simplify=None,
            instructions=None,
            instructions_format=None,
            roundabout_exits=None,
            attributes=None,
            radiuses=None,
            maneuvers=None,
            bearings=None,
            continue_straight=None,
            elevation=None,
            extra_info=None,
            suppress_warnings=None,
            options=None,
            dry_run=None
    ):
        """Get directions between an origin point and a destination point.

        For more information, visit https://openrouteservice.org/dev/#/api-docs/v2/directions/{profile}/post

        :param locations: The coordinates tuple the route should be calculated
            from in order of visit.
        :type locations: list of list

        :param profile: Specifies the mode of transport to use when calculating
            directions. One of ["driving-car", "driving-hgv", "foot-walking",
            "foot-hiking", "cycling-regular", "cycling-road",
            "cycling-mountain", "cycling-electric",]. Default "driving-car".
        :type profile: str

        :param format: Specifies the response format. One of ['json', 'geojson']. Default "json".
            Geometry format for "json" is Google's encodedpolyline.
        :type format: str

        :param preference: Specifies the routing preference. One of ["fastest, "shortest",
            "recommended"]. Default fastest.
        :type preference: str

        :param alternative_routes: Specifies whether alternative routes are computed, and parameters
            for the algorithm determining suitable alternatives. Must contain "share_factor", "target_count"
            and "weight_factor".
        :type alternative_routes: dict

        :param units: Specifies the distance unit. One of ["m", "km", "mi"]. Default "m".
        :type units: str

        :param language: Language for routing instructions. One of ["en", "de", "cn",
            "es", "ru", "dk", "fr", "it", "nl", "br", "se", "tr", "gr"].
        :type language: str

        :param geometry: Specifies whether geometry should be returned. Default True.
        :type geometry: bool

        :param geometry_simplify: Specifies whether to simplify the geometry.
            Default False.
        :type geometry_simplify: bool

        :param instructions: Specifies whether to return turn-by-turn instructions.
            Default True.
        :type instructions: bool

        :param instructions_format: Specifies the the output format for instructions.
            One of ["text", "html"]. Default "text".
        :type instructions_format: str

        :param roundabout_exits: Provides bearings of the entrance and all passed
            roundabout exits. Adds the 'exit_bearings' array to the 'step' object
            in the response. Default False.
        :type roundabout_exits: bool

        :param attributes: Returns route attributes on ["avgspeed", "detourfactor", "percentage"].
            Must be a list of strings. Default None.
        :type attributes: list of str

        :param maneuvers: Specifies whether the maneuver object is included into the step object or not. Default: False.
        :type maneuvers bool

        :param radiuses: A list of maximum distances (measured in
            meters) that limit the search of nearby road segments to every given waypoint.
            The values must be greater than 0, the value of -1 specifies no limit in
            the search. The number of radiuses must correspond to the number of waypoints.
            Default 50 km (ORS backend).
        :type radiuses: list of int

        :param bearings: Specifies a list of pairs (bearings and
            deviations) to filter the segments of the road network a waypoint can
            snap to. For example bearings=[[45,10],[120,20]]. Each pair is a
            comma-separated list that can consist of one or two float values, where
            the first value is the bearing and the second one is the allowed deviation
            from the bearing. The bearing can take values between 0 and 360 clockwise
            from true north. If the deviation is not set, then the default value of
            100 degrees is used. The number of pairs must correspond to the number
            of waypoints. Setting optimized=false is mandatory for this feature to
            work for all profiles. The number of bearings corresponds to the length
            of waypoints-1 or waypoints. If the bearing information for the last waypoint
            is given, then this will control the sector from which the destination
            waypoint may be reached.
        :type bearings: list of list

        :param continue_straight: Forces the route to keep going straight at waypoints not
            restricting u-turns even if u-turns would be faster. Default False.
        :type continue_straight: bool

        :param elevation: Specifies whether to return elevation values for points.
            Default False.
        :type elevation: bool

        :param extra_info: Returns additional information on ["steepness", "suitability",
            "surface", "waycategory", "waytype", "tollways", "traildifficulty", "roadaccessrestrictions"].
            Must be a list of strings. Default None.
        :type extra_info: list of str

        :param suppress_warnings: Tells the system to not return any warning messages in extra_info.
        :type suppress_warnings: bool

        :param options: Refer to https://openrouteservice.org/dev/#/api-docs/v2/directions/{profile}/geojson/post for
            detailed documentation. Construct your own dict() options object and paste it to your code.
        :type options: dict

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: A route from provided coordinates and restrictions.
        :rtype: :class:`routingpy.direction.Direction`
        """

        params = {"coordinates": locations}

        if preference:
            params["preference"] = preference

        if alternative_routes:
            if not isinstance(alternative_routes, dict):
                raise TypeError("alternative_routes must be a dict.")
            if not all([key in alternative_routes.keys() for key in ['share_factor', 'target_count', 'weight_factor']]):
                raise ValueError("alternative_routes needs 'share_factor', 'target_count', 'weight_factor' keys")
            params['alternative_routes'] = alternative_routes

        if units:
            params["units"] = units

        if language:
            params["language"] = language

        if geometry is not None:
            params["geometry"] = geometry

        if geometry_simplify is not None:
            params["geometry_simplify"] = geometry_simplify

        if instructions is not None:
            params["instructions"] = instructions

        if instructions_format:
            params["instructions_format"] = instructions_format

        if roundabout_exits is not None:
            params["roundabout_exits"] = roundabout_exits

        if attributes:
            params["attributes"] = attributes

        if radiuses:
            params["radiuses"] = radiuses

        if maneuvers is not None:
            params['maneuvers'] = maneuvers

        if bearings:
            params["bearings"] = bearings

        if continue_straight is not None:
            params["continue_straight"] = continue_straight

        if elevation is not None:
            params["elevation"] = elevation

        if extra_info:
            params["extra_info"] = extra_info

        if suppress_warnings is not None:
            params['suppress_warnings'] = suppress_warnings

        if options:
            if profile == 'driving-hgv' and options.get('profile_params'):
                if options['profile_params'].get('restrictions') and not options.get('vehicle_type'):
                    raise ValueError(
                        "ORS: options.vehicle_type must be specified for driving-hgv if restrictions are set."
                    )
            params['options'] = options

        return self._parse_direction_json(
            self._request(
                "/v2/directions/" + profile + '/' + format,
                get_params={},
                post_params=params,
                dry_run=dry_run
            ), format, units, alternative_routes
        )

    @staticmethod
    def _parse_direction_json(response, format, units, alternative_routes):
        if response is None:  # pragma: no cover
            return Direction()

        units_factor = 1
        if units == 'mi':
            units_factor = 0.621371 * 1000
        elif units == 'km':
            units_factor = 1000


        if format == 'geojson':
            if alternative_routes:
                routes = []
                for route in response['features']:
                    routes.append(
                        Direction(
                            geometry=route['geometry']['coordinates'],
                            distance=int(route['properties']['summary']['distance']),
                            duration=int(route['properties']['summary']['duration']),
                            raw=route
                        )
                    )
                return Directions(routes, response)
            else:
                geometry = response['features'][0]['geometry']['coordinates']
                duration = int(response['features'][0]['properties']['summary']['duration'])
                distance = int(response['features'][0]['properties']['summary']['distance'])
                return Direction(geometry=geometry, duration=duration, distance=distance, raw=response)
        elif format == 'json':
            if alternative_routes:
                routes = []
                for route in response['routes']:
                    geometry = [
                        list(reversed(coord))
                        for coord in utils.decode_polyline5(route['geometry'])
                    ]
                    routes.append(
                        Direction(
                            geometry=geometry,
                            distance=int(route['summary']['distance']),
                            duration=int(route['summary']['duration'] * units_factor),
                            raw=route
                        )
                    )
                return Directions(routes, response)
            else:
                geometry = [
                    list(reversed(coord))
                    for coord in utils.decode_polyline5(response['routes'][0]['geometry'])
                ]
                duration = int(response['routes'][0]['summary']['duration'])
                distance = int(response['routes'][0]['summary']['distance'] * units_factor)

                return Direction(geometry=geometry, duration=duration, distance=distance, raw=response)

    def isochrones(
            self,
            locations,
            profile,
            intervals,
            interval_type=None,
            units=None,
            location_type=None,
            smoothing=None,
            attributes=None,
            intersections=None,
            dry_run=None
    ):
        """Gets isochrones or equidistants for a range of time/distance values around a given set of coordinates.

        :param locations: One pair of lng/lat values.
        :type locations: list of float

        :param profile: Specifies the mode of transport to use when calculating
            directions. One of ["driving-car", "driving-hgv", "foot-walking",
            "foot-hiking", "cycling-regular", "cycling-safe", "cycling-mountain",
            "cycling-electric",]. Default "driving-car".
        :type profile: str

        :param interval_type: Set 'time' for isochrones or 'distance' for equidistants.
            Default 'time'.
        :type interval_type: str

        :param intervals: Ranges to calculate distances/durations for. This can be
            a list of multiple ranges, e.g. [600, 1200, 1400]. In meters or seconds.
        :type intervals: list of int

        :param units: Specifies the unit system to use when displaying results.
            One of ["m", "km", "m"]. Default "m".
        :type units: str

        :param location_type: 'start' treats the location(s) as starting point,
            'destination' as goal. Default 'start'.
        :type location_type: str

        :param smoothing: Applies a level of generalisation to the isochrone polygons generated.
            Value between 0 and 1, whereas a value closer to 1 will result in a more generalised shape.
        :type smoothing: float

        :param attributes: 'area' returns the area of each polygon in its feature
            properties. 'reachfactor' returns a reachability score between 0 and 1.
            'total_pop' returns population statistics from https://ghsl.jrc.ec.europa.eu/about.php.
            One or more of ['area', 'reachfactor', 'total_pop']. Default 'area'.
        :type attributes: list of str

        :param intersections: Write smth.
        :type intersections: bool

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: An isochrone with the specified range.
        :rtype: :class:`routingpy.isochrone.Isochrones`
        """

        params = {
            "locations": [locations],
            "profile": profile,
            "range": intervals,
        }

        if interval_type:
            params["range_type"] = interval_type

        if units:
            params["units"] = units

        if location_type:
            params["location_type"] = location_type

        if smoothing:
            params["smoothing"] = smoothing

        if attributes:
            params["attributes"] = attributes

        if intersections:
            params["intersections"] = intersections

        return self._parse_isochrone_json(
            self._request(
                "/v2/isochrones/" + profile + '/geojson',
                get_params={},
                post_params=params,
                dry_run=dry_run
            )
        )

    @staticmethod
    def _parse_isochrone_json(response):
        if response is None:  # pragma: no cover
            return Isochrones()

        isochrones = []
        for idx, isochrone in enumerate(response['features']):
            isochrones.append(
                Isochrone(
                    geometry=isochrone['geometry']['coordinates'][0],
                    interval=isochrone['properties']['value'],
                    center=isochrone['properties']['center']
                )
            )

        return Isochrones(isochrones=isochrones, raw=response)

    def matrix(
            self,
            locations,
            profile,
            sources=None,
            destinations=None,
            metrics=None,
            resolve_locations=None,
            units=None,
            dry_run=None
    ):
        """ Gets travel distance and time for a matrix of origins and destinations.

        :param locations: Two or more pairs of lng/lat values.
        :type locations: list of list

        :param profile: Specifies the mode of transport to use when calculating
            directions. One of ["driving-car", "driving-hgv", "foot-walking",
            "foot-hiking", "cycling-regular", "cycling-road", "cycling-mountain",
            "cycling-electric",]. Default "driving-car".
        :type profile: str

        :param sources: A list of indices that refer to the list of locations
            (starting with 0). If not passed, all indices are considered.
        :type sources: list of int

        :param destinations: A list of indices that refer to the list of locations
            (starting with 0). If not passed, all indices are considered.
        :type destinations: list of int

        :param metrics: Specifies a list of returned metrics. One or more of ["distance",
            "duration"]. Default ['duration'].
        :type metrics: list of str

        :param resolve_locations: Specifies whether given locations are resolved or
            not. If set 'true', every element in destinations and sources will
            contain the name element that identifies the name of the closest street.
            Default False.
        :type resolve_locations: bool

        :param units: Specifies the unit system to use when displaying results.
            One of ["m", "km", "m"]. Default "m".
        :type units: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: A matrix from the specified sources and destinations.
        :rtype: :class:`routingpy.matrix.Matrix`
        """

        params = {"locations": locations, "profile": profile}

        if sources:
            params['sources'] = sources

        if destinations:
            params['destinations'] = destinations

        if metrics:
            params["metrics"] = metrics

        if resolve_locations is not None:
            params["resolve_locations"] = resolve_locations

        if units:
            params["units"] = units

        return self._parse_matrix_json(
            self._request(
                "/v2/matrix/" + profile + '/json', get_params={}, post_params=params, dry_run=dry_run
            )
        )

    @staticmethod
    def _parse_matrix_json(response):
        if response is None:  # pragma: no cover
            return Matrix()
        durations = response.get('durations')
        distances = response.get('distances')
        return Matrix(durations=durations, distances=distances, raw=response)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""
:class:`Isochrone` returns directions results.
"""


class Isochrones(object):
    """
    Contains a list of :class:`Isochrone`, which can be iterated over or accessed by index. The property ¸`raw`` contains
    the complete raw response of the isochrones request.
    """

    def __init__(self, isochrones=None, raw=None):
        self._isochrones = isochrones
        self._raw = raw

    @property
    def raw(self):
        """
        Returns the isochrones's raw, unparsed response. For details, consult the routing engine's API documentation.

        :rtype: dict or None
        """
        return self._raw

    def __repr__(self):  # pragma: no cover
        return 'Isochrones({}, {})'.format(self._isochrones, self.raw)

    def __getitem__(self, item):
        return self._isochrones[item]

    def __iter__(self):
        return iter(self._isochrones)

    def __len__(self):
        return len(self._isochrones)


class Isochrone(object):
    """
    Contains a parsed single isochrone response. Access via properties ``geometry``, ``interval`` ``center``.
    """

    def __init__(self, geometry=None, interval=None, center=None):
        self._geometry = geometry
        self._interval = int(interval)
        self._center = center

    @property
    def geometry(self):
        """
        The geometry of the isochrone as [[lon1, lat1], [lon2, lat2], ...] list.

        :rtype: list or None
        """
        return self._geometry

    @property
    def center(self):
        """
        The center coordinate in [lon, lat] of the isochrone. Might deviate from the input coordinate.
        Not available for all routing engines (e.g. GraphHopper, Mapbox OSRM or Valhalla).
        In this case, it will use the location from the user input.

        :rtype: list of float
        """
        return self._center

    @property
    def interval(self):
        """
        The interval of the isochrone in seconds or in meters.

        :return: int
        """
        return self._interval

    def __repr__(self):  # pragma: no cover
        return 'Isochrone({}, {})'.format(self.geometry, self.interval)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""
:class:`Matrix` returns directions results.
"""


class Matrix(object):
    """
    Contains a parsed matrix response. Access via properties ``geometry`` and ``raw``.
    """

    def __init__(self, durations=None, distances=None, raw=None):
        self._durations = durations
        self._distances = distances
        self._raw = raw

    @property
    def durations(self):
        """
        The durations matrix as list akin to::

            [
                [
                    duration(origin1-destination1),
                    duration(origin1-destination2),
                    duration[origin1-destination3),
                    ...
                ],
                [
                    duration(origin2-destination1),
                    duration(origin2-destination2),
                    duration[origin3-destination3),
                    ...
                },
                ...
            ]

        :rtype: list or None
        """
        return self._durations

    @property
    def distances(self):
        """
        The distance matrix as list akin to::

            [
                [
                    duration(origin1-destination1),
                    duration(origin1-destination2),
                    duration[origin1-destination3),
                    ...
                ],
                [
                    duration(origin2-destination1),
                    duration(origin2-destination2),
                    duration[origin3-destination3),
                    ...
                },
                ...
            ]

        :rtype: list or None
        """
        return self._distances

    @property
    def raw(self):
        """
        Returns the matrices raw, unparsed response. For details, consult the routing engine's API documentation.

        :rtype: dict or None
        """
        return self._raw

    def __repr__(self):  # pragma: no cover
        return 'Matrix({}, {})'.format(self.durations, self.distances)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""
Defines exceptions that are thrown by the ORS client.
"""


class RouterError(Exception):  # pragma: no cover
    """Represents an exception returned by the remote or local API."""

    def __init__(self, status, message=None):
        self.status = status
        self.message = message

    def __str__(self):
        if self.message is None:
            return self.status
        else:
            return "%s (%s)" % (self.status, self.message)


class RouterApiError(RouterError):
    """Represents an exception returned by a routing engine, i.e. 400 <= HTTP status code <= 500"""


class RouterServerError(RouterError):
    """Represents an exception returned by a server, i.e. 500 <= HTTP"""


class RouterNotFound(Exception):
    """Represents an exception raised when router can not be found by name."""


class Timeout(Exception):  # pragma: no cover
    """The request timed out."""
    pass


class JSONParseError(Exception):  # pragma: no cover
    """The Json response can't be parsed.."""
    pass


class RetriableRequest(Exception):  # pragma: no cover
    """Signifies that the request can be retried."""
    pass


class OverQueryLimit(RouterError, RetriableRequest):
    """Signifies that the request failed because the client exceeded its query rate limit.

    Normally we treat this as a retriable condition, but we allow the calling code to specify that these requests should
    not be retried.
    """
    pass



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""
Core client functionality, common across all API requests.
"""

from .base import Router, DEFAULT
from routingpy import utils
from routingpy.direction import Direction
from routingpy.isochrone import Isochrone, Isochrones
from routingpy.matrix import Matrix

from operator import itemgetter


class Valhalla(Router):
    """Performs requests to a Valhalla instance."""

    def __init__(
            self,
            base_url,
            api_key=None,
            user_agent=None,
            timeout=DEFAULT,
            retry_timeout=None,
            requests_kwargs=None,
            retry_over_query_limit=False,
            skip_api_error=None
    ):
        """
        Initializes a Valhalla client.

        :param api_key: Mapbox API key. Required if base_url='https://api.mapbox.com/valhalla/v1'.
        :type api_key: str

        :param base_url: The base URL for the request. Defaults to the ORS API
            server. Should not have a trailing slash.
        :type base_url: str

        :param user_agent: User Agent to be used when requesting.
            Default :attr:`routingpy.routers.options.default_user_agent`.
        :type user_agent: str

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify ``None`` for no timeout. Default :attr:`routingpy.routers.options.default_timeout`.
        :type timeout: int or None

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.  Default :attr:`routingpy.routers.options.default_retry_timeout`.
        :type retry_timeout: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. **Note**, that ``proxies`` can be set globally
            in :attr:`routingpy.routers.options.default_proxies`.

            Example:

            >>> from routingpy.routers import Valhalla
            >>> router = Valhalla(my_key, requests_kwargs={'proxies': {'https': '129.125.12.0'}})
            >>> print(router.proxies)
            {'https': '129.125.12.0'}
        :type requests_kwargs: dict

        :param retry_over_query_limit: If True, client will not raise an exception
            on HTTP 429, but instead jitter a sleeping timer to pause between
            requests until HTTP 200 or retry_timeout is reached.
            Default :attr:`routingpy.routers.options.default_over_query_limit`.
        :type retry_over_query_limit: bool

        :param skip_api_error: Continue with batch processing if a :class:`routingpy.exceptions.RouterApiError` is
            encountered (e.g. no route found). If False, processing will discontinue and raise an error.
            Default :attr:`routingpy.routers.options.default_skip_api_error`.
        :type skip_api_error: bool
        """

        self.api_key = api_key

        super(Valhalla, self).__init__(
            base_url, user_agent, timeout, retry_timeout, requests_kwargs, retry_over_query_limit,
            skip_api_error
        )

    class Waypoint(object):
        """
        Constructs a waypoint with additional information, such as via or encoded lines.

        Example:

        >>> waypoint = Valhalla.WayPoint(position=[8.15315, 52.53151], type='break', heading=120, heading_tolerance=10, minimum_reachability=10, radius=400)
        >>> route = Valhalla('http://localhost/v1').directions(locations=[[[8.58232, 51.57234]], waypoint, [7.15315, 53.632415]])
        """

        def __init__(
                self,
                position,
                type=None,
                heading=None,
                heading_tolerance=None,
                minimum_reachability=None,
                radius=None,
                rank_candidates=None
        ):
            """

            :param type: Type of location. One of ['break', 'through']. A break is a stop, so the first
            and last locations must be of type break. A through location is one that the route path travels
            through, and is useful to force a route to go through location. The path is not allowed to
            reverse direction at the through locations. If no type is provided, the type is assumed to be a break.
            The order has to correspond to ``coordinates`` and be of the same length.
            More info at: https://github.com/valhalla/valhalla/blob/master/docs/api/turn-by-turn/api-reference.md#locations
            :type type: list of str

            :param heading: Preferred direction of travel for the start from the location. The heading is indicated
                in degrees from north in a clockwise direction, where north is 0°, east is 90°,
                south is 180°, and west is 270°.
            :type heading: list of int

            :param heading_tolerance: How close in degrees a given street's angle must be in order for it
                to be considered as in the same direction of the heading parameter. The default value is 60 degrees.
            :type heading_tolerance: list of int

            :param minimum_reachability: Minimum number of nodes (intersections) reachable for a given edge (road between
                intersections) to consider that edge as belonging to a connected region. When correlating this
                location to the route network, try to find candidates who are reachable from this many or more
                nodes (intersections). If a given candidate edge reaches less than this number of nodes its considered
                to be a disconnected island and we'll search for more candidates until we find at least one that
                isn't considered a disconnected island. If this value is larger than the configured service limit
                it will be clamped to that limit. The default is a minimum of 50 reachable nodes.
            :type minimum_reachability: list of int

            :param radius: The number of meters about this input location within which edges (roads between intersections)
                will be considered as candidates for said location. When correlating this location to the route network,
                try to only return results within this distance (meters) from this location. If there are no candidates
                within this distance it will return the closest candidate within reason. If this value is larger than
                the configured service limit it will be clamped to that limit. The default is 0 meters.
            :type radius: list of int

            :param rank_candidates: Whether or not to rank the edge candidates for this location. The ranking is used
                as a penalty within the routing algorithm so that some edges will be penalized more heavily than others.
                If true candidates will be ranked according to their distance from the input and various other attributes.
                If false the candidates will all be treated as equal which should lead to routes that are just the most
                optimal path with emphasis about which edges were selected.
            :type rank_candidates: list of bool
            """

            self._position = position
            self._type = type
            self._heading = heading
            self._heading_tolerance = heading_tolerance
            self._minimum_reachability = minimum_reachability
            self._radius = radius
            self._rank_candidates = rank_candidates

        def _make_waypoint(self):

            waypoint = {'lon': self._position[0], 'lat': self._position[1]}

            if self._type:
                waypoint['type'] = self._type

            if self._heading:
                waypoint['heading'] = self._heading

            if self._heading_tolerance:
                waypoint['heading_tolerance'] = self._heading_tolerance

            if self._minimum_reachability:
                waypoint['minimum_reachability'] = self._minimum_reachability

            if self._radius:
                waypoint['radius'] = self._radius

            if self._rank_candidates is not None:
                waypoint['rank_candidates'] = self._rank_candidates

            return waypoint

    def directions(
            self,
            locations,
            profile,
            options=None,
            units=None,
            language=None,
            directions_type=None,
            avoid_locations=None,
            date_time=None,
            id=None,
            dry_run=None
    ):
        """Get directions between an origin point and a destination point.

        For more information, visit https://github.com/valhalla/valhalla/blob/master/docs/api/turn-by-turn/api-reference.md.

        :param locations: The coordinates tuple the route should be calculated
            from in order of visit. Can be a list/tuple of [lon, lat] or :class:`Valhalla.WayPoint` instance or
            a combination of both.
        :type locations: list of list or list of :class:`Valhalla.WayPoint`

        :param profile: Specifies the mode of transport to use when calculating
            directions. One of ["auto", "auto_shorter", "bicycle", "bus", "hov", "motor_scooter",
            "motorcycle", "multimodal", "pedestrian".
        :type profile: str

        :param options: Profiles can have several options that can be adjusted to develop the route path,
            as well as for estimating time along the path. Only specify the actual options dict, the profile
            will be filled automatically. For more information, visit:
            https://github.com/valhalla/valhalla/blob/master/docs/api/turn-by-turn/api-reference.md#costing-options
        :type options: dict

        :param units: Distance units for output. One of ['mi', 'km']. Default km.
        :type units: str

        :param language: The language of the narration instructions based on the IETF BCP 47 language tag string.
            One of ['ca', 'cs', 'de', 'en', 'pirate', 'es', 'fr', 'hi', 'it', 'pt', 'ru', 'sl', 'sv']. Default 'en'.
        :type language: str

        :param directions_type: 'none': no instructions are returned. 'maneuvers': only maneuvers are returned.
            'instructions': maneuvers with instructions are returned. Default 'instructions'.
        :type directions_type: str

        :param avoid_locations: A set of locations to exclude or avoid within a route.
            Specified as a list of coordinates, similar to coordinates object.
        :type avoid_locations: list of list or list of :class:`Valhalla.WayPoint`

        :param date_time: This is the local date and time at the location. Field ``type``: 0: Current departure time,
            1: Specified departure time. Field ``value```: the date and time is specified
            in ISO 8601 format (YYYY-MM-DDThh:mm), local time.
            E.g. date_time = {type: 0, value: 2019-03-03T08:06:23}
        :type date_time: dict

        :param id: Name your route request. If id is specified, the naming will be sent thru to the response.
        :type id: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: A route from provided coordinates and restrictions.
        :rtype: :class:`routingpy.direction.Direction`
        """

        params = dict(costing=profile)

        params['locations'] = self._build_locations(locations)

        if options:
            params['costing_options'] = dict()
            profile = profile if profile != 'multimodal' else 'transit'
            params['costing_options'][profile] = options

        if any((units, language, directions_type)):
            params['directions_options'] = dict()
            if units:
                params['directions_options']['units'] = units
            if language:
                params['directions_options']['language'] = language
            if directions_type:
                params['directions_options']['directions_type'] = directions_type

        if avoid_locations:
            params['avoid_locations'] = self._build_locations(avoid_locations)

        if date_time:
            params['date_time'] = date_time

        if id:
            params['id'] = id

        get_params = {'access_token': self.api_key} if self.api_key else {}

        return self._parse_direction_json(
            self._request("/route", get_params=get_params, post_params=params, dry_run=dry_run), units
        )

    @staticmethod
    def _parse_direction_json(response, units):
        if response is None:  # pragma: no cover
            return Direction()

        geometry, duration, distance = [], 0, 0
        for leg in response['trip']['legs']:
            geometry.extend([list(reversed(coord)) for coord in utils.decode_polyline6(leg['shape'])])
            duration += leg['summary']['time']

            factor = 0.621371 if units == 'mi' else 1
            distance += int(leg['summary']['length'] * 1000 * factor)

        return Direction(geometry=geometry, duration=int(duration), distance=int(distance), raw=response)

    def isochrones(
            self,
            locations,
            profile,
            intervals,
            colors=None,
            polygons=None,
            denoise=None,
            generalize=None,
            options=None,
            units=None,
            language=None,
            directions_type=None,
            avoid_locations=None,
            date_time=None,
            show_locations=None,
            id=None,
            dry_run=None
    ):
        """Gets isochrones or equidistants for a range of time values around a given set of coordinates.

        For more information, visit https://github.com/valhalla/valhalla/blob/master/docs/api/isochrone/api-reference.md.

        :param locations: One pair of lng/lat values. Takes the form [Longitude, Latitude].
        :type locations: list of float

        :param profile: Specifies the mode of transport to use when calculating
            directions. One of ["auto", "bicycle", "multimodal", "pedestrian".
        :type profile: str

        :param intervals: Time ranges to calculate isochrones for. Up to 4 ranges are possible. In seconds.
        :type intervals: list of int

        :param colors: The color for the output of the contour. Specify it as a Hex value, but without the #, such as
            "color":"ff0000" for red. If no color is specified, the isochrone service will assign a default color to the output.
        :type colors: list of str

        :param polygons: Controls whether polygons or linestrings are returned in GeoJSON geometry. Default False.
        :type polygons: bool

        :param denoise: Can be used to remove smaller contours. In range [0, 1]. A value of 1 will only return the largest contour
            for a given time value. A value of 0.5 drops any contours that are less than half the area of the largest
            contour in the set of contours for that same time value. Default 1.
        :type denoise: float

        :param generalize: A floating point value in meters used as the tolerance for Douglas-Peucker generalization.
            Note: Generalization of contours can lead to self-intersections, as well as intersections of adjacent contours.
        :type generalize: float

        :param options: Profiles can have several options that can be adjusted to develop the route path,
            as well as for estimating time along the path. Only specify the actual options dict, the profile
            will be filled automatically. For more information, visit:
            https://github.com/valhalla/valhalla/blob/master/docs/api/turn-by-turn/api-reference.md#costing-options
        :type options: dict

        :param units: Distance units for output. One of ['mi', 'km']. Default km.
        :type units: str

        :param language: The language of the narration instructions based on the IETF BCP 47 language tag string.
            One of ['ca', 'cs', 'de', 'en', 'pirate', 'es', 'fr', 'hi', 'it', 'pt', 'ru', 'sl', 'sv']. Default 'en'.
        :type language: str

        :param avoid_locations: A set of locations to exclude or avoid within a route.
            Specified as a list of coordinates, similar to coordinates object.
        :type avoid_locations: list of list

        :param date_time: This is the local date and time at the location. Field ``type``: 0: Current departure time,
            1: Specified departure time. Field ``value```: the date and time is specified
            in format YYYY-MM-DDThh:mm, local time.

            E.g. date_time = {type: 0, value: 2019-03-03T08:06}
        :type date_time: dict

        :param id: Name your route request. If id is specified, the naming will be sent thru to the response.
        :type id: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: An isochrone with the specified range.
        :rtype: :class:`routingpy.isochrone.Isochrones`
        """

        location_objects = self._build_locations(locations)

        contours = []
        for idx, r in enumerate(intervals):
            d = {'time': int(r / 60)}
            if colors:
                try:
                    d.update(color=colors[idx])
                except IndexError:
                    raise IndexError("Colors object must have same length as Range object.")
            contours.append(d)

        params = {
            "locations": location_objects,
            "costing": profile,
            "contours": contours,
        }

        if options:
            params['costing_options'] = dict()
            profile = profile if profile != 'multimodal' else 'transit'
            params['costing_options'][profile] = options

        if polygons is not None:
            params['polygons'] = polygons

        if denoise:
            params['denoise'] = denoise

        if generalize:
            params['generalize'] = generalize

        if avoid_locations:
            params['avoid_locations'] = self._build_locations(avoid_locations)

        if date_time:
            params['date_time'] = date_time

        if show_locations is not None:
            params['show_locations'] = show_locations

        if id:
            params['id'] = id

        get_params = {'access_token': self.api_key} if self.api_key else {}
        return self._parse_isochrone_json(
            self._request("/isochrone", get_params=get_params, post_params=params, dry_run=dry_run),
            intervals, locations
        )

    @staticmethod
    def _parse_isochrone_json(response, intervals, locations):
        if response is None:  # pragma: no cover
            return Isochrones()

        isochrones = []
        for idx, feature in enumerate(reversed(response['features'])):
            if feature['geometry']['type'] in ('LineString', 'Polygon'):
                isochrones.append(Isochrone(
                    geometry=feature['geometry']['coordinates'],
                    interval=intervals[idx],
                    center=locations
                ))

        return Isochrones(isochrones, response)

    def matrix(
            self,
            locations,
            profile,
            sources=None,
            destinations=None,
            options=None,
            avoid_locations=None,
            units=None,
            id=None,
            dry_run=None
    ):
        """ Gets travel distance and time for a matrix of origins and destinations.

        :param locations: Multiple pairs of lng/lat values.
        :type locations: list of list

        :param profile: Specifies the mode of transport to use when calculating
            matrices. One of ["auto", "bicycle", "multimodal", "pedestrian".
        :type profile: str

        :param sources: A list of indices that refer to the list of locations
            (starting with 0). If not passed, all indices are considered.
        :type sources: list of int

        :param destinations: A list of indices that refer to the list of locations
            (starting with 0). If not passed, all indices are considered.
        :type destinations: list of int

        :param options: Profiles can have several options that can be adjusted to develop the route path,
            as well as for estimating time along the path. Only specify the actual options dict, the profile
            will be filled automatically. For more information, visit:
            https://github.com/valhalla/valhalla/blob/master/docs/api/turn-by-turn/api-reference.md#costing-options
        :type options: dict

        :param avoid_locations: A set of locations to exclude or avoid within a route.
            Specified as a list of coordinates, similar to coordinates object.
        :type avoid_locations: list of list

        :param units: Distance units for output. One of ['mi', 'km']. Default km.
        :type units: str

        :param id: Name your route request. If id is specified, the naming will be sent through to the response.
        :type id: str

        :param dry_run: Print URL and parameters without sending the request.
        :param dry_run: bool

        :returns: A matrix from the specified sources and destinations.
        :rtype: :class:`routingpy.matrix.Matrix`
            """

        params = {
            'costing': profile,
        }

        locations = self._build_locations(locations)

        sources_coords = locations
        if sources is not None:
            sources_coords = itemgetter(*sources)(sources_coords)
            if isinstance(sources_coords, dict):
                sources_coords = [sources_coords]
        params['sources'] = sources_coords

        dest_coords = locations
        if destinations is not None:
            dest_coords = itemgetter(*destinations)(dest_coords)
            if isinstance(dest_coords, dict):
                dest_coords = [dest_coords]
        params['targets'] = dest_coords

        if options:
            params['costing_options'] = dict()
            profile = profile if profile != 'multimodal' else 'transit'
            params['costing_options'][profile] = options

        if avoid_locations:
            params['avoid_locations'] = self._build_locations(avoid_locations)

        if units:
            params["units"] = units

        if id:
            params['id'] = id

        get_params = {'access_token': self.api_key} if self.api_key else {}

        return self._parse_matrix_json(
            self._request(
                '/sources_to_targets', get_params=get_params, post_params=params, dry_run=dry_run
            ), units
        )

    @staticmethod
    def _parse_matrix_json(response, units):
        if response is None:  # pragma: no cover
            return Matrix()

        factor = 0.621371 if units == 'mi' else 1
        durations = [
            [destination['time'] for destination in origin] for origin in response['sources_to_targets']
        ]
        distances = [
            [int(destination['distance'] * 1000 * factor)
             for destination in origin]
            for origin in response['sources_to_targets']
        ]

        return Matrix(durations=durations, distances=distances, raw=response)

    def _build_locations(self, coordinates):
        """Build the locations object for all methods"""

        locations = []

        # Isochrones only support one coordinate tuple, so check for type of first element
        if isinstance(coordinates[0], (list, tuple, self.Waypoint)):
            for idx, coord in enumerate(coordinates):
                if isinstance(coord, (list, tuple)):
                    locations.append({'lon': coord[0], 'lat': coord[1]}),
                elif isinstance(coord, self.Waypoint):
                    locations.append(coord._make_waypoint())
                else:
                    raise TypeError(
                        "Location type {} at index {} is not supported: {}".format(
                            type(coord), idx, coord
                        )
                    )
        elif isinstance(coordinates[0], float):
            locations.append({'lon': coordinates[0], 'lat': coordinates[1]})

        return locations



# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from .valhalla import Valhalla
from .base import DEFAULT


class MapboxValhalla(Valhalla):
    """Performs requests to Mapbox's Valhalla instance."""

    _base_url = 'https://api.mapbox.com/valhalla/v1'

    def __init__(
            self,
            api_key,
            user_agent=None,
            timeout=DEFAULT,
            retry_timeout=None,
            requests_kwargs=None,
            retry_over_query_limit=False,
            skip_api_error=None
    ):
        """
        Initializes a Valhalla client.

        :param api_key: Mapbox API key.
        :type api_key: str

        :param user_agent: User Agent to be used when requesting.
            Default :attr:`routingpy.routers.options.default_user_agent`.
        :type user_agent: str

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify ``None`` for no timeout. Default :attr:`routingpy.routers.options.default_timeout`.
        :type timeout: int or None

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.  Default :attr:`routingpy.routers.options.default_retry_timeout`.
        :type retry_timeout: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. **Note**, that ``proxies`` can be set globally
            in :attr:`routingpy.routers.options.default_proxies`.

            Example:

            >>> from routingpy.routers import MapboxValhalla
            >>> router = MapboxValhalla(my_key, requests_kwargs={'proxies': {'https': '129.125.12.0'}})
            >>> print(router.proxies)
            {'https': '129.125.12.0'}
        :type requests_kwargs: dict

        :param retry_over_query_limit: If True, client will not raise an exception
            on HTTP 429, but instead jitter a sleeping timer to pause between
            requests until HTTP 200 or retry_timeout is reached.
            Default :attr:`routingpy.routers.options.default_over_query_limit`.
        :type retry_over_query_limit: bool

        :param skip_api_error: Continue with batch processing if a :class:`routingpy.exceptions.RouterApiError` is
            encountered (e.g. no route found). If False, processing will discontinue and raise an error.
            Default :attr:`routingpy.routers.options.default_skip_api_error`.
        :type skip_api_error: bool
        """

        super(MapboxValhalla, self).__init__(
            self._base_url, api_key, user_agent, timeout, retry_timeout, requests_kwargs,
            retry_over_query_limit, skip_api_error
        )
